#!/usr/bin/env python

# Attach features (columns) to CoNLL-formatted data.

from __future__ import with_statement

import sys
import re

options = None

# "empty" missing feature value used for padding for tokens for which
# there are fewer than the maximum number of features.
MISSING_FEATURE = 'O'

def argparser():
    import argparse
    
    ap=argparse.ArgumentParser(description="Attach features to " \
                                   "CoNLL-formatted data.")
    ap.add_argument("-t", "--tokenindex", default="3", 
                    help="Index of field with tokens (default 3)")
    ap.add_argument("-c", "--count", default=None, 
                    help="Number of features to add")
    ap.add_argument("-v", "--verbose", default=False, action="store_true", 
                    help="Verbose output")    
    ap.add_argument("features", metavar="FEATURES", nargs=1, 
                    help="Feature file (\"-\" for STDIN)")
    ap.add_argument("data", metavar="DATA", nargs='+', 
                    help="CoNLL-formatted data")
    return ap

##### CoNLL format parsing #####

class ParseError(Exception):
    def __init__(self, line, linenum, message=None, filename=None):
        self.line = line
        self.linenum = linenum
        self.message = message
        self.file = file

        if self.message is None:
            self.message = "Parse error"

    def __str__(self):
        return (self.message +
                ("on line %d" % self.linenum) + 
                ("" if self.file is None else " in file %s" % self.file) +
                (": '%s'" % self.line))

EMPTY_LINE_RE = re.compile(r'^\s*$')

def _line_is_empty(l):
    return EMPTY_LINE_RE.match(l)

def parse_conll(input, filename=None, separator='\t', is_empty=_line_is_empty):
    """Parse CoNLL-style input.

    Input should consist of blocks of lines separated by empty lines
    (is_empty), each non-empty line consisting of fields separated by
    the given separator.

    Returns:
        list of lists of lists: blocks, lines, fields.
    """

    li, l = 0, None
    try:
        blocks = []
        current_block = []
        for l in input:
            l = l.rstrip()
            li += 1
            if is_empty(l):
                blocks.append(current_block)
                current_block = []
            else:
                current_block.append(l.split(separator))
    except Exception:
        # whatever goes wrong
        raise ParseError(l, li)

    return blocks

##### End CoNLL format parsing #####

def read_features(f):
    feature_map = {}

    # simple tab-separated format: first field is token, other fields
    # are features for the token.

    for l in f:
        l = l.rstrip('\n')
        fields = l.split('\t')        
        feature_map[fields[0]] = fields[1:]

    return feature_map

def process(fn, feature_map, feature_num):
    global options
    
    with open(fn, 'rU') as f:
        blocks = parse_conll(f, fn)

    for block in blocks:
        for fields in block:
            token = fields[options.tokenindex]

            feats = feature_map.get(token, [])

            # pad if necessary
            padcount = feature_num - len(feats)
            if padcount > 0:
                feats.extend([MISSING_FEATURE for i in range(padcount)])

            fields.extend(feats)

            print '\t'.join(fields)
        print

def main(argv=None):
    if argv is None:
        argv = sys.argv

    global options
    options = argparser().parse_args(argv[1:])

    # index parameters
    try:
        options.tokenindex = int(options.tokenindex)
    except Exception:
        print >> sys.stderr, "Error: failed to parse '%s' as index" % options.tokenindex
        argparser().print_help()
        return 1        

    # read features
    if options.features[0] == "-":
        feature_map = read_features(sys.stdin)
    else:
        with open(options.features[0], 'rU') as f:
            feature_map = read_features(f)            
    max_feature_num = max([len(f) for f in feature_map.values()] + [0])

    if options.count is None:
        feature_num = max_feature_num
    else:
        feature_num = max(max_feature_num, int(options.count))

    # primary processing
    for fn in options.data:
        process(fn, feature_map, feature_num)

    return 0

if __name__ == "__main__":
    sys.exit(main(sys.argv))
