import re

BRACE_RE = re.compile(r'^(.*?)\{([^{}]+)\}(.*)$')

def expand_braces(s):
    """Perform bash-like brace expansion on string.

    For example,
    >>> expand_braces('O:{1,2,3}')
    ['O:1', 'O:2', 'O:3']

    >>> expand_braces('x{1..5}')
    ['x1', 'x2', 'x3', 'x4', 'x5']

    >>> expand_braces('x{1..5..2},y{2,4}')
    ['x1,y2', 'x1,y4', 'x3,y2', 'x3,y4', 'x5,y2', 'x5,y4']

    >>> expand_braces('{3..0..-1}')
    ['3', '2', '1', '0']

    >>> expand_braces('{1.0..3.0..0.5}')
    ['1.0', '1.5', '2.0', '2.5', '3.0', '3.5']
    
    Note: the implementation does not replicate the full bash brace
    expansion algorithm and is likely to behave differently also for
    some implemented cases.
    """
    
    m = BRACE_RE.match(s)
    if not m:
        return [s]
    else:
        pre, braced, post = m.groups()
        # combine and recurse
        expanded = [expand_braces(pre+e+post) for e in 
                    _expand_brace_expression(braced)]
        # return flattened (http://stackoverflow.com/a/952952)
        return [e for nested in expanded for e in nested]

# TODO: character ranges        
RANGE_EXP_RE = re.compile(r'^(-?\d+(?:\.\d+)?)\.\.(-?\d+(?:\.\d+)?)(?:\.\.(-?\d+(?:\.\d+)?))?$')

# http://stackoverflow.com/a/7267280
def frange(val, end, incr):
  if val < end and incr > 0:
      while val < end:
          yield val
          val += incr
  elif val > end and incr < 0:
      while val > end:
          yield val
          val += incr
  else:
      raise ValueError

def _expand_brace_expression(s):
    m = RANGE_EXP_RE.match(s)
    if m:
        start, end, incr = m.groups()
        # int if possible, float otherwise
        try:
            start, end = int(start), int(end)
            incr = int(incr) if incr is not None else 1
        except ValueError:
            start, end = float(start), float(end)
            incr = float(incr) if incr is not None else 1
        # inclusive range
        end = end+(1 if end > start else -1)
        return [str(i) for i in frange(start, end, incr)]
    elif ',' in s:
        return s.split(',')
    else:
        return [s]

def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()
