import json
import os

from types import StringTypes

class ConfigError(Exception):
    pass

def _loads(s):
    if not s or s.isspace():
        return {}
    else:
        return json.loads(s)

def _dumps(o):
    if isinstance(o, StringTypes) and (not o or o.isspace):
        return '{}'
    else:
        return json.dumps(o, sort_keys=True, indent=2)

class Config(object):
    """JSON-based storage for configuration settings.  

    Can be accessed using either dictionary or attribute syntax
    (conf['foo'] or conf.foo). Values default to None if not present,
    and None values are not stored.
    """

    def __init__(self, filename=None):
        self.__dict__['_filename'] = filename # avoid __setattr__
        self.load()

    def _load(self):
        # TODO: error handling
        try:
            with open(self._filename, 'rt') as f:
                return _loads(f.read())
        except IOError, e:
            raise ConfigError(str(e))
        except ValueError, e:
            raise ConfigError('%s: %s' % (str(e), self._filename))

    def load(self):
        if self._filename is None or not os.path.exists(self._filename):
            conf = {}
        else:
            conf = self._load()
        self.__dict__['_config'] = conf
        return self._config

    def save(self):
        if self._config is None:
            raise ConfigError('no config to save')
        if self._filename is None:
            raise ConfigError('no config filename')
        try:
            with open(self._filename, 'wt') as f:
                f.write(str(self))
        except IOError, e:
            raise ConfigError(str(e))

    def __str__(self):
        try:
            return _dumps(self._config)
        except ValueError, e:
            raise ConfigError(str(e))
        except TypeError, e:
            raise ConfigError(str(e))

    def update(self, key, value, default=None):
        """Set key to value.  

        If value is None and the Config does not already have the key,
        use default instead."""

        if value is not None:
            self[key] = value
        elif self[key] is None:
            self[key] = default

    def __getitem__(self, key):
        return self._config.get(key, None)

    def __setitem__(self, key, value):
        if value is not None:
            self._config[key] = value
        else:
            try:
                del self._config[key]
            except KeyError:
                pass

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value
            
if __name__ == '__main__':
    from tempfile import mkstemp

    # test / examples
    tmp_fh, tmp_path = mkstemp()
    os.close(tmp_fh)

    config = Config(tmp_path)
    print config
    config['foo'] = 1
    print config
    print config['foo']
    print config.foo
    print config['bar']
    print config.bar
    config.save()
    
    os.remove(tmp_path)
