from types import StringTypes

from tools import pairwise
import input

class Processor(object):
    """Abstract base class for document processors.

    Some document processors such as ML-based taggers may have
    non-trivial startup costs and require resource deallocation on
    exit. To assure proper resource management and support flexible
    creation of processors, __init__() should only configure the
    processor, and resources should be allocated on start() and
    released on stop(). __enter__() and __exit__() are provided, and
    usage should follow patterns such as

    with FooProcessor() as processor:
        for doc in documents:
            processor.process(doc)    

    or invocation via process_documents().
    """

    def __init__(self):
        """Configure processor."""
        self.running = False

    def process(self, document):
        """Process given document."""
        assert self.running, 'error: process() on stopped Processor'
        self._process(document)

    def start(self):
        """Start processor, reserving necessary resources."""
        assert not self.running, 'error: start() on running Processor'
        self.running = True

    def stop(self):
        """Stop processor, releasing resources."""
        assert self.running, 'error: stop() on stopped Processor'
        self.running = False

    def _process(self, document):
        """Implementation for process()"""
        raise NotImplementedError
    
    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()

def process_documents(documents, processors, callback=None):
    """Process documents with processors, invoking callback after
    processing each.

    documents must be an iterable containing either Document objects
    or strings. The latter will be interpreted as the names of files
    from which to read the input Documents.

    The processors will be start()ed prior to the beginning of
    processing and stop()ped after tagging the last file.
    """

    def cb(running):
        _process_documents(documents, running, callback)

    return run_processors(processors, cb)

def _process_documents(documents, running, callback):
    """process_documents() implementation.

    Do not invoke directly.
    """

    for document in documents:
        try:
            if isinstance(document, StringTypes):
                # assume filename
                document = input.read(document)
            for processor in running:
                processor.process(document)
            if callback is not None:
                callback(document)
        except Exception:
            import sys
            if isinstance(document, StringTypes):
                docid = document
            else:
                docid = document.id
            print >> sys.stderr, 'Error processing document', docid
            raise

    return True

def run_processors(processors, callback=None):
    """Start processors and invoke callback.

    The callback is provided with list of running processors as its
    only argument.

    The processors will be start()ed prior to invoking the callback
    and stop()ped after the callback returns.
    """
    return _run_processors(processors, [], callback)

def _run_processors(pending, running, callback):
    """run_processors() implementation.

    Do not invoke directly.
    """

    # If there are pending processors, start one and recurse;
    # otherwise invoke callback. Recursive implementation to allow
    # with statements to be used with variable numbers of processors.

    if pending:
        with pending[0] as processor:
            _run_processors(pending[1:], running+[processor], callback)
    else:
        if callback is not None:
            return callback(running)
        else:
            return None

# TODO: these should go someplace else, such as a general utilities.py

from unicode2ascii import read_u2a_data, convert_u2a

class ASCIIConverter(Processor):
    def start(self):
        Processor.start(self)
        read_u2a_data()
    
    def _process(self, document):
        # traverse sentences, store sentence head (text before the
        # first token), then traverse tokens in pairs, storing the
        # tail of each token (space between the token and the next),
        # convert token text to ASCII using convert_u2a(), redo token
        # offsets, then redo the sentence text using the head, token
        # texts and tails, and finally redo the document text using
        # the sentences.

        document.unicode_text = document.text
        offset = 0
        for sentence in document.sentences:
            sentence.unicode_text = sentence.text
            sent_end = sentence.base_offset + len(sentence.text)
            ascii_base_offset = offset

            # sentence head is 
            if sentence.tokens:
                t_start = sentence.tokens[0].start
                sent_head = document.text[sentence.base_offset:t_start]
            else:
                sent_head = sentence.text
            assert not sent_head or sent_head.isspace()

            offset += len(sent_head)

            for t, next_t in pairwise(sentence.tokens, include_last=True):
                t.unicode_text = t.text
                tail_end = next_t.start if next_t is not None else sent_end
                t.tail = document.text[t.end:tail_end]
                assert not t.tail or t.tail.isspace()
                t.text = convert_u2a(t.text)
                t.start, t.end = offset, offset + len(t.text)
                offset += len(t.text) + len(t.tail)

            sentence.base_offset = ascii_base_offset
            sentence.text = sent_head + ''.join([t.text + t.tail 
                                                 for t in sentence.tokens])

            assert sentence.is_valid()
        
        document.text = ''.join(s.text for s in document.sentences)
        assert document.is_valid()

class DocumentStorage(Processor):
    """Stores processed documents."""

    def __init__(self, storage):
        Processor.__init__(self)
        self.storage = storage

    def _process(self, document):
        self.storage.append(document)

from tagsequence import OUT_TAG, parse_tag, make_tag

class SingleTyper(Processor):
    """Replaces all annotation types with a single type."""

    def __init__(self, type_='span'):
        Processor.__init__(self)
        self.type = type_

    def _process(self, document):
        for sentence in document.sentences:
            for token in sentence.tokens:
                if token.tag != OUT_TAG:
                    token.tag = make_tag(parse_tag(token.tag)[0], self.type)

from tagsequence import DEFAULT_TAGSET, BIOES_TAGSET, BIO_to_BIOES

class BIOESConverter(Processor):
    """Convert token BIO tag sequences to BIOES."""

    def _process(self, document):
        for sentence in document.sentences:
            for t, next_t in pairwise(sentence.tokens, include_last=True):
                next_tag = next_t.tag if next_t is not None else None
                t.tag = BIO_to_BIOES(t.tag, next_tag)
            
def tagset_converter(tagset=None):
    if tagset is None:
        tagset = DEFAULT_TAGSET
    if tagset == DEFAULT_TAGSET:
        return []
    elif tagset == BIOES_TAGSET:
        return [BIOESConverter()]
    else:
        assert False, 'unexpected tagset value %s' % tagset
