import sys
import os
import re

import random
import tempfile
import shutil

import common

def argparser():
    ap = common.argparser(eval_options=True)

    # TODO: why aren't the following set in common.argparser()?
    ap.add_argument('-t', '--test', metavar='FILE', nargs='+', default=None,
                    help='test data')
    ap.add_argument('-T', '--tmpdir', default='tmp', metavar='DIR',
                    help='directory for temporary data')
    ap.add_argument('-x', '--cross-validate', metavar='N', default=None,
                    type=int,
                    help='perform N-fold cross-validation on training data')
    return ap

def preprocess(documents):
    global options

    processed = []

    processors = (common.input_processors(options) +
                  common.tagging_processors(options) +
                  [common.DocumentStorage(processed)])

    common.process_documents(documents, processors)

    return processed

def evaluate(train_documents, test_documents, c2_values=None, b_values=None,
             output=None):
    global options

    if output is None:
        output = sys.stdout
    if c2_values is None:
        c2_values = [1.0]
    if b_values is None:
        b_values = [None]

    tags = set()
    for d in train_documents:
        tags |= d.get_tags()

    for c2 in c2_values:
        # c2 changes require retraining, b changes do not.
        options.c2 = c2
        processors = [common.NERsuiteTrainer.from_CLI_options(options)]
        common.process_documents(train_documents, processors)
        for b in b_values:
            # wildcard expansion in b arguments
            options.b = common.expand_b_argument(b, tags)
            evaltagger = common.NERsuiteEvalTagger.from_CLI_options(options)
            processors = [evaltagger]
            if options.output is not None:
                processors.append(common.PredictionWriter(options.output,
                                                          options.format,
                                                          options.encoding))
            common.process_documents(test_documents, processors)
            prefix = "c2 %f\tb %s\t" % (c2, b)
            print >> output, common.summarize_results(evaltagger.evaluators,
                                                      prefix)
            output.flush()

def cross_validate(documents, N, c2_values=None, b_values=None, 
                   output=None):
    global options

    random.seed()
    documents = documents[:]
    random.shuffle(documents)

    # each fold is a (train, test) pair where both are lists of
    # documents
    folds = [([],[]) for n in range(N)]
    for d in range(len(documents)):
        for n in range(N):
            if d % N != n:
                folds[n][0].append(documents[d])
            else:
                folds[n][1].append(documents[d])

    for n, fold in enumerate(folds):
        train, test = fold
        evaluate(train, test, c2_values, b_values, output)

def verify_options(options, prog=None, err=None):
    if prog is None:
        pref = 'error:'
    else:
        pref = prog + ': error: '
    if err is None:
        err = sys.stderr

    if options.test is None and options.cross_validate is None:
        print >> err, pref + 'argument -t or -x required'
        return False
    if options.test is not None and options.cross_validate is not None:
        print >> err, pref + 'both -t and -x not possible'
        return False
    if options.cross_validate is not None and options.cross_validate < 2:
        print >> err, pref + 'argument -x: value should be > 1'
        return False

    return True

def main(argv=None):
    global options

    if argv is None:
        argv = sys.argv
    options = argparser().parse_args(argv[1:])
    # TODO: use config system here

    if not verify_options(options, argv[0]):
        return 1

    # if the user hasn't specified a model, pick a temporary one that
    # will be deleted on exit
    delete_model = False
    if options.model is None:
        # assure temp directory exists, ignore errors (e.g. EEXIST)
        try:
            os.makedirs(options.tmpdir)
        except OSError:
            pass
        options.model = tempfile.mkdtemp(suffix='.model', dir=options.tmpdir)
        delete_model = True
    else:
        if not os.path.exists(options.model):
            os.mkdir(options.model)
        elif not os.path.isdir(options.model):
            assert False, '%s is not a model directory' % options.model

    # determine parameter values to evaluate
    c2_values = []
    for c2 in  common.parameter_values(options.c2):
        try:
            c2_values.append(float(c2))
        except ValueError:
            print >> sys.stderr, 'error: argument -c2: invalid float value', c2
            return 1
    b_values = common.parameter_values(options.b, default=[None])

    # select documents and short-circuit on missing
    selected_train = [d for d in common.select_documents(options.files, True)]
    if not selected_train:
        print >> sys.stderr, 'error: no training documents found'
        return 1
    if options.test is None:
        selected_test = []
    else:
        selected_test = [d for d in common.select_documents(options.test, True)]
        if not selected_test:
            print >> sys.stderr, 'error: test no documents found'
            return 1

    # pre-process and store documents to avoid repeating processing in
    # parameter selection    
    train_documents = preprocess(selected_train)
    if selected_test:
        test_documents = preprocess(selected_test)
    else:
        test_documents = []

    try:
        if options.test is not None:
            evaluate(train_documents, test_documents, c2_values, b_values)
        else:
            cross_validate(train_documents, options.cross_validate,
                           c2_values, b_values)
    finally:
        if delete_model:
            shutil.rmtree(options.model)

    return 0
