#!/usr/bin/env python

from collections import defaultdict
from itertools import izip

from tagsequence import TypedSpan, parse_tag, make_tag, tagged_spans

def match_exact(span, gold):
    """Return true if TypedSpans match exactly, False otherwise."""

    return span == gold

def match_left(span, gold):
    """Return true if TypedSpans match in type and start token (left
    boundary match), False otherwise."""

    return span.type == gold.type and span.start == gold.start

def match_right(span, gold):
    """Return true if TypedSpans match in type and end token (right
    boundary match), False otherwise."""

    return span.type == gold.type and span.end == gold.end

def match_overlap(span, gold):
    """Return true if TypedSpans match in type and overlap on at least
    one token, False otherwise."""

    return span.type == gold.type and (span.end > gold.start and
                                       span.start < gold.end)

class Evaluator(object):
    """Encapsulates matching criteria and a set of evaluation results."""

    def __init__(self, criteria=match_exact, results=None):
        if results is None:
            results = EvalResult()
        self.match = criteria
        self.results = results

    def evaluate_spans(self, spans, gold_spans):
        """Evaluate TypedSpan list against gold TypedSpan list."""

        # TODO: avoid N^2
#        matched_span = { s : False for s in spans }
#        matched_gold = { s : False for s in gold_spans }
        matched_span = dict([(s, False) for s in spans])
        matched_gold = dict([(s, False) for s in gold_spans])
        for s in spans:
            for g in gold_spans:
                if self.match(s, g):
                    matched_span[s] = True
                    matched_gold[g] = True
        for s in gold_spans:
            if matched_gold[s]:
                self.results.add_true_positive_gold(s.type)
            else:
                self.results.add_false_negative(s.type)
        for s in spans:
            if matched_span[s]:
                self.results.add_true_positive_pred(s.type)
            else:
                self.results.add_false_positive(s.type)

    def criterion(self):
        return self.criterion_name(self.match)

    def __str__(self):
        return '%s: %s' % (self.criterion(), self.results)

    @classmethod
    def criterion_name(cls, match_function):
        return match_function.__name__.replace('match_', '')


def default_evaluators():
    return [Evaluator(match_exact), Evaluator(match_left),
            Evaluator(match_right), Evaluator(match_overlap)]

def evaluate_tags(tags, gold_tags, evaluators=None, single_type=None):
    """Evaluate tags against gold tags and return evaluators.

    Apply given Evaluators if not None, otherwise create new default
    evaluators.

    If single_type is not None, treat all tagged spans as having the
    given type, ignoring differences in tagged types.
    """

    if evaluators is None:
        evaluators = default_evaluators()

    assert len(tags) == len(gold_tags)

    if single_type:
        tags, gold_tags = retype_tags(tags), retype_tags(gold_tags)

    spans, gold_spans = tagged_spans(tags), tagged_spans(gold_tags)

    for e in evaluators:
        e.evaluate_spans(spans, gold_spans)

    return evaluators
                
def retype_tags(tags, type_):
    """Return tags with all types replaced with given single type."""
    
    single_typed = []
    for t in tags:
        ttag, ttype = parse_tag(t)
        single_typed.append(make_tag(ttype, None) if ttype is None else
                            make_tag(ttype, type_))
    return single_typed

def summarize_results(evaluators, prefix=''):
    """Return a string summarizing the results of the given Evaluators
    by type. The prefix will be prepended to each line in the summary."""

    criteria = [e.criterion() for e in evaluators]
    summaries = [e.results.per_type_summary() for e in evaluators]
    types = [s[0] for s in summaries[0]]

    summary = []
    for i, type_ in enumerate(types):
        summary.append(prefix + type_)
        for j, criterion in enumerate(criteria):
            assert summaries[j][i][0] == type_, 'type mismatch'
            summary.append('\t' + criterion + ': ' + summaries[j][i][1])
        summary.append('\n')

    return ''.join(summary)    

class EvalResult(object):
    """Evaluation results for some matching criterion."""

    def __init__(self):
        # NOTE: separating true positives for predictions from true
        # positives for gold to account for a prediction matching many
        # things in gold (or vice versa) for approximate matching
        # criteria.
        self.true_positive_pred = defaultdict(int)
        self.true_positive_gold = defaultdict(int)
        self.false_positive = defaultdict(int)
        self.false_negative = defaultdict(int)
        self.total_true_positive_pred = 0
        self.total_true_positive_gold = 0
        self.total_false_positive = 0
        self.total_false_negative = 0

    def get_true_positive_pred(self, type_=None):
        """Return number of true positive predictions of given type.

        If type is None, return total number of true positives.
        """

        if type_ is None:
            return self.total_true_positive_pred
        else:
            return self.true_positive_pred[type_]

    def get_true_positive_gold(self, type_=None):
        """Return number of true positive gold of given type.

        If type is None, return total number of true positives.
        """

        if type_ is None:
            return self.total_true_positive_gold
        else:
            return self.true_positive_gold[type_]

    def get_false_positive(self, type_=None):
        """Return number of false positives of given type.

        If type is None, return total number of false positives.
        """

        if type_ is None:
            return self.total_false_positive
        else:
            return self.false_positive[type_]

    def get_false_negative(self, type_=None):
        """Return number of false negatives of given type.

        If type is None, return total number of false negatives.
        """

        if type_ is None:
            return self.total_false_negative
        else:
            return self.false_negative[type_]

    def add_true_positive_pred(self, type_=None, count=1):
        """Add count to true positive predictions of given type."""

        assert count > 0
        self.true_positive_pred[type_] += count
        self.total_true_positive_pred += count

    def add_true_positive_gold(self, type_=None, count=1):
        """Add count to true positive gold of given type."""

        assert count > 0
        self.true_positive_gold[type_] += count
        self.total_true_positive_gold += count

    def add_false_positive(self, type_=None, count=1):
        """Add count to false positives of given type."""

        assert count > 0
        self.false_positive[type_] += count
        self.total_false_positive += count

    def add_false_negative(self, type_=None, count=1):
        """Add count to false negatives of given type."""

        assert count > 0
        self.false_negative[type_] += count
        self.total_false_negative += count

    def precision(self, type_=None, NaN=None):
        """Return precision for given type, or overall precision
        (microaverage) if None.

        If precision is not a number (TP + FP == 0), return NaN if not
        None, otherwise raise ValueError.
        """

        TP, FP = self.get_true_positive_pred(type_), \
            self.get_false_positive(type_)
        if TP + FP != 0:
            return 1.*TP/(TP+FP)
        elif NaN is not None:
            return NaN
        else:
            raise ValueError

    def recall(self, type_=None, NaN=None):
        """Return recall for given type, or overall recall
        (microaverage) if None.

        If recall is not a number (TP + FN == 0), return NaN if not
        None, otherwise raise ValueError.
        """

        TP, FN = self.get_true_positive_gold(type_), \
            self.get_false_negative(type_)
        if TP + FN != 0:
            return 1.*TP/(TP+FN)
        elif NaN is not None:
            return NaN
        else:
            raise ValueError

    def f_score(self, type_=None, NaN=None):
        """Return F-score for given type, or overall F-score
        (microaverage) if None.

        If F-score is not a number (precision == recall == 0), return
        NaN if not None, otherwise raise ValueError.
        """

        prec, rec = self.precision(type_, 0.0), self.recall(type_, 0.0)

        if prec + rec != 0.0:
            return 2*prec*rec/(prec+rec)
        elif NaN is not None:
            return NaN
        else:
            raise ValueError

    def precision_str(self, type_=None, NaN=0.0):
        prec = self.precision(type_, NaN)
        TP, FP = self.get_true_positive_pred(type_), \
            self.get_false_positive(type_)
        total = TP + FP
        return '%.2f%% (%d/%d)' % (100.*prec, TP, total)

    def recall_str(self, type_=None, NaN=0.0):
        rec = self.recall(type_, NaN)
        TP, FN = self.get_true_positive_gold(type_), \
            self.get_false_negative(type_)
        total = TP + FN
        return '%.2f%% (%d/%d)' % (100.*rec, TP, total)

    def f_score_str(self, type_=None, NaN=0.0):
        f = self.f_score(type_, NaN)
        return '%.2f%%' % (100.*f)

    def summary_str(self, type_=None, NaN=0.0):
        return 'p %s r %s f %s' % (self.precision_str(type_, NaN),
                                   self.recall_str(type_, NaN),
                                   self.f_score_str(type_, NaN))

    def per_type_summary(self, NaN=0.0, total_str="TOTAL"):
        """Return list of (type, result) pairs for each type and a
        total, where result is a p/r/f summary string."""
        
        types = set(sum([d.keys() for d in (self.true_positive_pred,
                                            self.true_positive_gold,
                                            self.false_positive,
                                            self.false_negative)], []))        
        result = []
        for type_ in sorted(list(types)) + [None]:
            result.append((type_ if type_ is not None else total_str,
                           self.summary_str(type_)))
        return result

    def __str__(self):
        return self.summary_str(None)
