import sys
import re
import os

import codecs

from collections import defaultdict

from document import Document
from tools import txt_for_ann
from pubmedxml import pubmed_xml_to_text

def select_documents(names, labeled_only=False):
    for n in names:
        if n == '-':
            for line in sys.stdin:
                yield Document.from_text(line)
        elif os.path.isfile(n):
            yield n            
        elif os.path.isdir(n):
            for m in select_directory(n, labeled_only):
                yield m
        else:
            print >> sys.stderr, "Don't know how to read \"%s\"" % n

_extension_priority = {
    '.nersuite': 4,
    '.ann' : 3,
    '.xml' : 2,
    '.txt' : 1,
}

def _select_extension(exts):
    return sorted(exts, lambda a,b: cmp(_extension_priority.get(b,-1),
                                        _extension_priority.get(a,-1)))[0]

def select_directory(name, labeled_only=False):
    files = [n for n in os.listdir(name) if can_read(n)]

    if labeled_only:
        files = [n for n in files if is_labeled(n)]

    # group by root (name w/o extension) and return only single name
    # for each root to avoid redundant processing and issues arising
    # e.g. from first converting a .ann and then its corresponding
    # .txt (without annotation).

    by_root = defaultdict(list)
    for n in files:
        root, ext = os.path.splitext(os.path.basename(n))
        by_root[root].append(ext)

    for root in by_root:
        choice = _select_extension(by_root[root])
        yield os.path.join(name, root+choice)    

def read_nersuite(filename, encoding):
    with codecs.open(filename, 'rU', encoding=encoding) as in_:
        return Document.from_nersuite(in_.read())

def read_txt(filename, encoding):
    with codecs.open(filename, 'rU', encoding=encoding) as in_:
        return Document.from_text(in_.read())

def read_ann(filename, encoding):
    txtfilename = txt_for_ann(filename)
    with codecs.open(txtfilename, 'rU', encoding=encoding) as t_in:
        with codecs.open(filename, 'rU', encoding=encoding) as a_in:
            return Document.from_standoff(t_in.read(), a_in.read())

def read_pubmed_xml(filename, encoding):
    # ElementTree wants a byte stream, not Unicode
    with open(filename) as in_:
        return Document.from_text(pubmed_xml_to_text(in_))

def read_xml(filename, encoding):
    # TODO: format recognition
    return read_pubmed_xml(filename, encoding)

_READER = {
    '.txt' : read_txt,
    '.ann' : read_ann,
    '.nersuite' : read_nersuite,
    '.xml' : read_xml,
}

def can_read(filename):
    return os.path.splitext(filename)[1] in _READER    

def is_labeled(filename):
    return os.path.splitext(filename)[1] != '.txt'

def read(filename, encoding='utf-8'):
    """Parse given file and return Document.

    If the given file requires supporting files (such as the text for
    standoff annotations), these will also be processed.
    """

    if filename == '-':
        return Document.from_text(sys.stdin.read())

    ext = os.path.splitext(filename)[1]

    if ext not in _READER:
        print >> sys.stderr, "Don't know how to read \"%s\"" % filename
        assert False # TODO: raise exception specific to the issue
    else:
        doc = _READER[ext](filename, encoding)
        doc.id = os.path.splitext(os.path.basename(filename))[0]
        return doc
