from document import FORMATS, DEFAULT_FORMAT
from docproc import SingleTyper, ASCIIConverter, tagset_converter
from tagger import GTagger, DictTagger
from casing import CASING_OPTIONS, DEFAULT_CASING, casing_processor
from tagsequence import TAGSETS, DEFAULT_TAGSET
from config import Config

USAGE='''usage: %s COMMAND [OPTIONS] FILE [FILE [...]]

COMMAND:
    learn   train tagger with given data
    tag     apply trained tagger to given data
    eval    train, apply and evaluate tagger
    convert convert from one format to another

OPTIONS:
    command-specific, use COMMAND -h for details'''

SHORT_USAGE='%(prog)s COMMAND [OPTIONS] FILE [FILE [...]]'

DEFAULT_ENCODING = 'UTF-8'
DEFAULT_MODEL = 'default.model'
DEFAULT_CONFIG = 'config.json'

options=None

def argparser(train_options=False, tag_options=False, eval_options=False,
              convert_options=False):
    """Return ArgumentParser with options common to all commands plus
    requested command-specific ones."""

    import argparse

    ap=argparse.ArgumentParser(description='Scientific Entity Mention suite',
                               usage=SHORT_USAGE)
    ap.add_argument('-1', '--singletype', default=None, metavar='TYPE',
                    help='replace all annotation types with TYPE')
    ap.add_argument('-a', '--asciify', default=None, action='store_true',
                    help='map input to ASCII')
    ap.add_argument('-c', '--casing', default=None, choices=CASING_OPTIONS,
                    help='input casing (default %s)' % DEFAULT_CASING)
    ap.add_argument('-d', '--dictmodels', default=None,
                    help='Dictionary models (semicolon-separated)')
    ap.add_argument('-e', '--encoding', default=None,
                    help='text encoding (default %s)' % DEFAULT_ENCODING)
    ap.add_argument('-f', '--format', metavar='FORMAT',
                    default=None, choices=FORMATS,
                    help='output format (default %s)' % DEFAULT_FORMAT)
    ap.add_argument('-m', '--model', default=None, metavar='FILE',
                    help='location for model (default %s)' % DEFAULT_MODEL)
    ap.add_argument('-s', '--tagset', choices=TAGSETS, default=None,
                    help='tagset (default %s)' % DEFAULT_TAGSET)
    ap.add_argument('-v', '--verbose', default=None, action='store_true', 
                    help='verbose output')
    ap.add_argument('files', metavar='FILE', nargs='+',
                    help='input data ("-" for STDIN)')

    # TODO: reorder so that these options appear in more reasonable
    # spots in the list.

    if train_options:
        ap.add_argument('-c2', default=None, type=float, metavar='FLOAT',
                        help='L2 regularization coefficient')
    elif eval_options:
        # more permissive type checking (i.e. none) for c ranges in eval
        ap.add_argument('-c2', default=None, metavar='FLOAT[,...]',
                        help='L2 regularization coefficient(s)')

    if train_options or tag_options or eval_options:
        ap.add_argument('-b', default=None, metavar='TAG:BIAS[,...]',
                        help='label bias(es) (e.g. "B-Gene:2.0,O:0.5")')

    if tag_options or eval_options or convert_options:
        ap.add_argument('-o', '--output', metavar='FILE/DIR', default=None,
                        help='file or directory to output to')        

    return ap

def _reset_model(path):
    import os
    import shutil
    # TODO: error handling
    shutil.rmtree(path, ignore_errors=True)
    os.mkdir(path)

def get_configuration(options, revise=True):
    import os.path

    if options.model is None:
        options.model = DEFAULT_MODEL
        if revise:
            _reset_model(options.model)
    if not os.path.exists(options.model):
        os.mkdir(options.model)
    elif not os.path.isdir(options.model):
        assert False, '%s is not a model directory' % options.model

    conf_path = os.path.join(options.model, DEFAULT_CONFIG)
    conf = Config(conf_path)

    # propagate non-default options, applying defaults
    # TODO: make this less brittle
    conf.update('singletype', options.singletype)
    conf.update('asciify', options.asciify)
    conf.update('casing', options.casing, DEFAULT_CASING)
    conf.update('dictmodels', options.dictmodels)
    conf.update('encoding', options.encoding, DEFAULT_ENCODING)
    conf.update('format', options.format, DEFAULT_FORMAT)
    conf.update('model', options.model, DEFAULT_MODEL)
    conf.update('tagset', options.tagset, DEFAULT_TAGSET)
    conf.update('verbose', options.verbose, False)
    # OK for the following to be missing
    for a in ['c2', 'b', 'output']:
        if a in vars(options):
            conf.update(a, vars(options)[a])

    if revise:
        conf.save()

    return conf

def input_processors(options):
    """Return list of processors that the options specify to apply to
    documents before primary processing."""
    
    processors = []
    
    processors.extend(casing_processor(options.casing))

    if options.asciify:
        processors.append(ASCIIConverter())

    if options.singletype is not None:
        processors.append(SingleTyper(options.singletype))

    processors.extend(tagset_converter(options.tagset))

    return processors

def tagging_processors(options):
    processors = []

    # always run GTagger for morphosyntactic analysis
    processors.append(GTagger())

    if options.dictmodels is not None:
        for d in options.dictmodels.split(';'):
            processors.append(DictTagger.from_CLI_argument(d))

    return processors

def output_processors(options):
    """Return list of processors that the options specify to apply to
    documents after primary processing."""

    processors = []

    # none defined yet.

    return processors
