#!/usr/bin/env python

import sys
import re

from os import path
from subprocess import Popen, PIPE

from docproc import Processor
from tagger import ExternalTagger
from tagsequence import fix_sentence_BIO
from evaluation import evaluate_tags
from tools import pairwise

NERSUITE_CMD = 'nersuite'
DEFAULT_MODEL = 'nersuite.model'

def _get_model_path(modeldir):
    if modeldir is not None:
        modelpath = path.join(modeldir, DEFAULT_MODEL)
    else:
        modelpath = DEFAULT_MODEL
    return modelpath
    
class NERsuiteTagger(ExternalTagger):
    def __init__(self, parameters=None, modeldir=None, log=sys.stderr):
        ExternalTagger.__init__(self)

        if parameters is None:
            parameters = []
        modelpath = _get_model_path(modeldir)

        self.parameters = parameters
        self.modelpath = modelpath
        self.log = log

    def start(self):
        ExternalTagger.start(self)
        cmd = [NERSUITE_CMD, 'tag', '-m', self.modelpath] + self.parameters
        self.verify_command(cmd)
        self.processes.append(Popen(cmd, stdin=PIPE, stdout=PIPE, bufsize=1))

    def _tag_sentence(self, sentence):
        """Run NERsuite to tag Sentence.

        Output is used to set Token tags. Errors in predicted BIO
        sequences are corrected heuristically.
        """

        input_ = ''.join(sentence.to_nersuite(exclude_tag=True)).encode('utf-8')
        output_lines = self._communicate(input_)
        self._set_tags(sentence, output_lines)
        fix_sentence_BIO(sentence)

    @classmethod
    def _set_tags(cls, sentence, tagger_output):
        # special case for empty "sentences"
        if not sentence.tokens:
            assert not tagger_output
            return

        # sanity
        assert tagger_output[-1] == '\n'
        assert len(sentence) == len(tagger_output)-1

        for i, token in enumerate(sentence.tokens):
            line = tagger_output[i].rstrip('\n')
            start, end, ttext, features = cls._parse_tagger_line(line)
            tag = line.split('\t')[-1]

            assert (token.start == start and token.end == end and 
                    token.text == ttext), 'input/output mismatch'
            
            cls._set_tag(token, tag)

    @classmethod
    def _set_tag(cls, token, tag):
        token.tag = tag

    @classmethod
    def from_CLI_options(cls, options):
        """Initialize and return NERsuiteTagger with settings based on
        command-line options."""

        parameters = []
        if options.b is not None:
            parameters.extend(['-b', options.b])
        return cls(parameters, options.model)

    @classmethod
    def verify_command(cls, cmd):
        # TODO: more and more precise constraints
        for a, next_a in pairwise(cmd):
            if a == '-b':
                assert ':' in next_a, 'invalid -b argument form "%s"' % next_a

class NERsuiteEvalTagger(NERsuiteTagger):
    def __init__(self, parameters=None, modeldir=None, log=sys.stderr):
        NERsuiteTagger.__init__(self, parameters, modeldir, log)
        self.evaluators = None
        
    def _tag_sentence(self, sentence):
        gold_tags = [t.tag for t in sentence.tokens]
        NERsuiteTagger._tag_sentence(self, sentence)
        tags = [t.predicted_tag for t in sentence.tokens]
        self.evaluators = evaluate_tags(tags, gold_tags, self.evaluators)

    @classmethod
    def _set_tag(cls, token, tag):
        token.predicted_tag = tag
        
class NERsuiteTrainer(Processor):
    def __init__(self, parameters=None, modeldir=None, log=sys.stderr):
        Processor.__init__(self)

        if parameters is None:
            parameters = []
        modelpath = _get_model_path(modeldir)

        self.parameters = parameters
        self.modelpath = modelpath
        self.log = sys.stderr
        self.input = None
        self.running = False

    def start(self):
        Processor.start(self)
        cmd = [NERSUITE_CMD, 'learn', '-m', self.modelpath] + self.parameters
        self.nersuite = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)

    def _process(self, document):
        self.nersuite.stdin.write(document.to_nersuite())

    def stop(self):
        Processor.stop(self)
        self.nersuite.stdin.close()
        for l in self.nersuite.stderr:
            _nersuite_log(l, self.log)
        self.nersuite.wait()

    @classmethod
    def from_CLI_options(cls, options):
        """Initialize and return NERsuiteTrainer with settings based
        on command-line options."""

        parameters = []
        if options.c2 is not None:
            parameters.extend(['-C', str(options.c2)])
        return cls(parameters, options.model)

def _nersuite_log(line, log):
    # compacted CRFSuite/NERsuite logging output
    if not line or line.isspace():
        return
    line = re.sub(r'\*+\sIteration\s(\#\d+)\s*\*+\n', r'It. \1 ', line)
    line = re.sub(r'^Loss: (.*)\n', r'Loss \1 ', line)
    line = re.sub(r'^Feature norm: (.*)\n', r'Fn \1 ', line)
    line = re.sub(r'^Error norm: (.*)\n', r'En \1 ', line)
    line = re.sub(r'^Active features: (.*)\n', r'F# \1 ', line)
    line = re.sub(r'^Line search trials: (.*)\n', r'Tri \1 ', line)
    line = re.sub(r'^Line search step: (.*)\n', r'Step \1 ', line)
    line = re.sub(r'^Seconds required .*: (.*)', r'Sec \1', line)
    log.write(line)
