import sys
import os

import codecs

from document import Document
from document import FORMATS
from document import TEXT_FORMAT, STANDOFF_FORMAT, NERSUITE_FORMAT, \
    BC2GM_FORMAT, DEFAULT_FORMAT
from tools import noopmanager, txt_for_ann
from types import StringTypes
from docproc import Processor
from main import DEFAULT_ENCODING

class DocumentWriter(Processor):
    def __init__(self, output, format, encoding=None):
        Processor.__init__(self)
        self.output = output
        self.format = format
        if encoding is None:
            encoding = DEFAULT_ENCODING
        self.encoding = encoding

    def _process(self, document):
        # TODO: adjust output by document name
        write(self.output, document, self.format, self.encoding)

    @classmethod
    def from_CLI_options(cls, options):
        """Initialize and return DocumentWriter with settings based on
        command-line options."""

        if options.output is None:
            out = sys.stdout
        else:
            out = options.output
        return cls(out, options.format, options.encoding)

class PredictionWriter(DocumentWriter):
    """DocumentWriter that outputs predicted tags."""

    def _process(self, document):
        for s in document.sentences:
            for t in s.tokens:
                t.gold_tag, t.tag = t.tag, t.predicted_tag
        DocumentWriter._process(self, document)
        for s in document.sentences:
            for t in s.tokens:
                t.tag = t.gold_tag

def _out_stream(output, docid, suffix, encoding):
    if isinstance(output, StringTypes):
        if not os.path.isdir(output):
            ofn = output            
        else:
            ofn = os.path.join(output, docid+'.'+suffix)
        return codecs.open(ofn, 'wt', encoding=encoding)
    else:
        # TODO: does this make a difference?
        return noopmanager(codecs.getwriter(encoding)(output))
        
def write_nersuite(output, document, encoding):
    with _out_stream(output, document.id, 'nersuite', encoding) as out:
        out.write(document.to_nersuite())

def write_txt(output, document, encoding):
    with _out_stream(output, document.id, 'txt', encoding) as out:
        out.write(document.text)

def write_ann(output, document, encoding):
    txtoutput = txt_for_ann(output) if isinstance(output, StringTypes) \
        else output
    with _out_stream(txtoutput, document.id, 'txt', encoding) as t_out:
        with _out_stream(output, document.id, 'ann', encoding) as a_out:
            t_out.write(document.text)
            a_out.write(document.to_standoff())

def write_bc2gm(output, document, encoding):
    txtoutput = txt_for_ann(output) if isinstance(output, StringTypes) \
        else output
    with _out_stream(txtoutput, document.id, 'in', encoding) as t_out:
        with _out_stream(output, document.id, 'out', encoding) as a_out:
            t_out.write(document.bc2gm_text())
            a_out.write(document.to_bc2gm())

_WRITER = {
    TEXT_FORMAT : write_txt,
    STANDOFF_FORMAT : write_ann,
    NERSUITE_FORMAT : write_nersuite,
    BC2GM_FORMAT : write_bc2gm,
}

def write(output, document, format, encoding='utf-8'):
    """Write document to output in specified format.

    If output is a string, it will be interpreted as the name of the
    directory or file to write to; otherwise it is assumed to be a
    file-like object and written to using write(). If the given format
    requires supporting files (such as the text for standoff
    annotations), these will also be written.
    """

    if format is None:
        format = DEFAULT_FORMAT
    
    assert format in FORMATS and format in _WRITER

    _WRITER[format](output, document, encoding)
