#!/usr/bin/env python

# PubMed XML -> text conversion and related functionality.

# From BioNLP Shared Task supporting resources, under the MIT license.

import sys
import re
import os
import gzip

try:
    import xml.etree.ElementTree as ET
except ImportError: 
    import cElementTree as ET   

from tools import FormatError

def onlychild(element, childtag):
    children = element.findall(childtag)    
    if len(children) != 1:
        raise FormatError('expected single %s child in <%s>, found %d' % 
                          (childtag, element.tag, len(children)))
    return children[0]

def pubmed_xml_to_text(fn, single_line_abstract=False):
    return '\n'.join(read_abstracts(fn, single_line_abstract))

def read_abstracts(fn, single_line_abstract=False):
    input = fn

    # if the extension suggests a zip, try to wrap
    gzipped = False
    try:
        if re.search('\.gz$', fn):
            input = gzip.GzipFile(fn)
            gzipped = True
    except TypeError:
        pass

    all_abstracts = []
        
    for event, element in ET.iterparse(input):
        if event != 'end' or element.tag != 'MedlineCitation':
            continue

        citation = element
        PMID = onlychild(citation, 'PMID')
        article = onlychild(citation, 'Article')
        articleTitle = onlychild(article, 'ArticleTitle')

        # Article typically (but not always) contains an Abstract
        abstracts = article.findall('Abstract')
        if len(abstracts) > 1:
            raise FormatError('%d abstracts for PMID %s' % \
                                  (len(abstracts, PMID.text)))
        abstract = abstracts[0] if abstracts else None

        # if there's no Abstract, try to look for <OtherAbstract> in
        # the citation (not the article) element, which seems to be
        # used in some cases. Take the first if several.
        if abstract is None:
            otherAbstracts = citation.findall('OtherAbstract')
            if otherAbstracts:
                abstract = otherAbstracts[0]

        abstractText = None
        textAbstract = ''
        if abstract is not None:
            # if there's an Abstract, it should contain an AbstractText
            abstractTexts = abstract.findall('AbstractText')
            if not abstractTexts:
                raise FormatError('%d abstract texts for PMID %s' % \
                                      (len(abstractTexts), PMID.text))

            if len(abstractTexts) == 1:
                abstractText = abstractTexts[0]
                textAbstract = abstractText.text
            else:
                # recent versions of PubMed data may contain multiple
                # AbstractText elements for structured abstracts. In these
                # cases, "label" attributes give structured abstract
                # section headers and should be combined into the text.
                assert len(abstractTexts) > 1, 'INTERNAL ERROR'

                sectionTexts = []

                for at in abstractTexts:
                    # there may be instances of empty <AbstractText>
                    # elements in the data (see e.g. PMID 20619000 in
                    # the PubMed 2012 baseline). Skip those with the
                    # special "empty" label "UNLABELLED" entirely;
                    # print the label only for the rest.
                    if ((at.text is None or at.text.isspace()) and 
                        at.attrib.get('Label', '') == 'UNLABELLED'):
                        continue

                    t = ''
                    if ('Label' in at.attrib and
                        at.attrib['Label'] != 'UNLABELLED'):
                        t = at.attrib['Label'] + ':'

                    if at.text and not at.text.isspace():
                        if t and not t.isspace():
                            t += ' ' if single_line_abstract else '\n'
                        t += at.text
                    sectionTexts.append(t)

                separator = ' ' if single_line_abstract else '\n'
                textAbstract = separator.join(sectionTexts)

        textPMID     = PMID.text
        textTitle    = articleTitle.text        

        parts = [textTitle]
        if textAbstract:
            parts.append(textAbstract)
        all_abstracts.append('\n'.join(parts)+'\n')

        citation.clear()

    # if we were wrapping a .gz, close the GzipFile
    if gzipped:
        input.close()

    return all_abstracts
