#!/usr/bin/env python

import sys
import json

from types import StringTypes

# TODO: looser coupling with Document
from document import Document

from so2html import standoff_to_html

try:
    import bottle
except ImportError:
    print >> sys.stderr, 'Failed to import bottle. Please check installation.'
    sys.exit(1)

DEVELOPMENT = True

if DEVELOPMENT:
    try:
        from werkzeug.debug import DebuggedApplication
        bottle.debug(True)
    except:
        print >> sys.stderr, 'Failed to set up Werkzeug.'

TAGGER_NAME = 'AnatomyTagger'
SERVER_HOST = 'localhost'
SERVER_PORT = 8152
BASE_PATH   = 'tag'
AJAX_PATH   = 'ajax'
TEXT_ARG    = 'text'
FORMAT_ARG  = 'format'

FORMATS = [
    'html',
    'json',
    'standoff',
    'nersuite',
    ]

# Max number of characters in input document (5000 should be enough
# for most scientific paper abstracts)
MAX_DOCUMENT_LENGTH = 5000

document_processors = None

app = bottle.Bottle()

def base_url():
    return '%s:%d/%s' % (SERVER_HOST, SERVER_PORT, BASE_PATH)

def ajax_url():
    return '%s:%d/%s' % (SERVER_HOST, SERVER_PORT, AJAX_PATH)

EXAMPLE_TEXT='''Heart, brain, liver, blood and lumen.'''

def so2dict(text, so):
    return { 'text': text, 'ann': [{ 'start': s.start, 'end': s.end, 
                                     'type': s.type } for s in so] }
def validate_text(text):
    if len(text) > MAX_DOCUMENT_LENGTH:
        bottle.abort(400, 'Document too large')

def tag(text):
    validate_text(text)

    if document_processors is None:
        return text  # debugging no-op
    else:
        doc = Document.from_text(text)
        for proc in document_processors:
            proc.process(doc)
        return doc

def plaintext_response():
    bottle.response.content_type = 'text/plain'
            
@app.post('/'+BASE_PATH)
def tag_base():
    text = bottle.request.params.get(TEXT_ARG)
    format = bottle.request.params.get(FORMAT_ARG)

    if not text:
        bottle.abort(400, 'No text')
    if not format:
        format = FORMATS[0]
    if format not in FORMATS:
        bottle.abort(400, 'Invalid format: %s' % format)

    tagged = tag(text)

    if format == 'html':
        return standoff_to_html(text, tagged.standoffs())
    elif format == 'json':
        return so2dict(text, tagged.standoffs())
    elif format == 'standoff':
        plaintext_response()
        return tagged.to_standoff()
    elif format == 'nersuite':
        plaintext_response()
        return tagged.to_nersuite()
    else:
        bottle.abort(500, 'Format handling')

@app.post('/'+AJAX_PATH)
def tag_json():
    data = bottle.request.body.read()
    if not data:
        bottle.abort(400, 'No data')

    try:
        obj = json.loads(data)
    except ValueError, e:
        bottle.abort(400, 'Invalid JSON: %s' % e)

    # accept either plain string or object (dict) with attribute 'text'
    if isinstance(obj, StringTypes):
        text = obj
    else:
        try:
            text = obj['text']
        except KeyError:
            bottle.abort(400, 'No text: %s' % data)

    #return json.dumps(so2dict(tag(text)), indent=2, separators=(',', ': '))
    return so2dict(text, tag(text).standoffs())

@app.get('/')
def index():
    return '''<!DOCTYPE html>
<html>
  <head>
  <title>%s</title>
  <style type="text/css">
    html {background-color: #eee; font-family: sans;}
    body {background-color: #fff; border: 1px solid #ddd;
        padding: 15px; margin: 15px;}
    pre {background-color: #eee; border: 1px solid #ddd; padding: 5px;}
  </style>
  <script type="text/javascript" src="/lib/js/jquery-1.9.1.js"></script>
</head>
<body>
  <h1>%s server</h1>

  <h2>Usage: HTTP POST</h2>
  <p>URL: %s</p>
  <p>Parameters:</p>
  <dl>
    <dt>text:</dt><dd>text to tag</dd>
    <dt>format:</dt><dd> "html", "json", "standoff" or "nersuite"</dd>
  </dl>
  <h2>Example:</h2>
  <form action="/%s" method="post">
    <textarea rows="8" cols="50" name="%s">%s</textarea><br>
    Output format: <select name="format">
      <option value="html">HTML</option>
      <option value="json">JSON</option>
      <option value="standoff">standoff</option>
      <option value="nersuite">NERsuite (CoNLL-style)</option>
    </select><br>
    <input type="submit" value="tag">
  </form>

  <h2>Usage: AJAX</h2>
  <p>URL: %s</p>
  <p>Parameters (pass as JSON object):</p>
  <dl>
    <dt>text:</dt><dd>text to tag</dd>
  </dl>
  <h2>Example</h2>
  <form action="#" method="post">
    <textarea id="jsonin" rows="8" cols="50">%s</textarea><br>
    <input type="submit" value="tag" onclick="
  document.getElementById('jsonout').value = '';
  jQuery.ajax({
    type: 'POST',
    url: '%s',
    data: JSON.stringify({ text: document.getElementById('jsonin').value }),
    success: function(response) {
      document.getElementById('jsonout').value = 
        JSON.stringify(response, undefined, 2);
    },
    error: function(response) {
      alert('failed:' + response.statusText);
      console.log('failed:', response);
    }
  });
  return false;
">
  </form>
  <textarea id="jsonout" rows="8" cols="50" disabled="disabled"></textarea>
</body>
</html>
''' % (TAGGER_NAME, TAGGER_NAME, base_url(), 
       BASE_PATH, TEXT_ARG, EXAMPLE_TEXT,
       ajax_url(), EXAMPLE_TEXT, AJAX_PATH)

@app.get('/lib/<path:path>')
def get_static(path):
    return bottle.static_file(path, 'lib')

def serve(processors=None, port=None):
    global app, document_processors, SERVER_PORT

    document_processors = processors

    if DEVELOPMENT:
        app.catchall = False
        bottle_app, app = app, DebuggedApplication(app)

    if port is not None:
        SERVER_PORT = port

    bottle.run(app, host=SERVER_HOST, port=SERVER_PORT, reloader=DEVELOPMENT)

    return 0

def main(argv=None):
    return serve()

if __name__ == '__main__':
    sys.exit(main(sys.argv))
