#!/usr/bin/env python

# Extract tag-based features from CoNLL-formatted data.

from __future__ import with_statement

import sys
import re

from collections import defaultdict

# token-tag pair frequency (two layers of defaultdict with int default value)
token_tag_freq = defaultdict(lambda: defaultdict(int))

# first tag per token
first_tag = {}

# default positions of tag and token text.
TAG_INDEX = 0
TOKEN_INDEX = 3

def argparser():
    import argparse
    ap=argparse.ArgumentParser(description="Extract tag-based features " \
                                   "from CoNLL-formatted data")
    ap.add_argument("-i", "--tagindex", default=TAG_INDEX, 
                    help="Index of field with tags (default %d%s)" % 
                    (TAG_INDEX, " for last" if TAG_INDEX == -1 else ""))
    ap.add_argument("-t", "--tokenindex", default=TOKEN_INDEX, 
                    help="Index of field with tokens (default %d)" % 
                    TOKEN_INDEX)
    ap.add_argument("-s", "--skipout", default=False, action="store_true",
                    help="Skip output for tokens that only have \"O\" tags.")
    ap.add_argument("-v", "--verbose", default=False, action="store_true",
                    help="Verbose output.")
    ap.add_argument("files", nargs='+',
                    help="Target file(s) (\"-\" for STDIN)")
    return ap

class ParseError(Exception):
    def __init__(self, line, linenum, message=None, filename=None):
        self.line = line
        self.linenum = linenum
        self.message = message
        self.file = file

        if self.message is None:
            self.message = "Parse error"

    def __str__(self):
        return (self.message +
                (" on line %d" % self.linenum) + 
                ("" if self.file is None else " in file %s" % self.file) +
                (": '%s'" % self.line))

def parse_BIO_tag(tag):
    """Parse given string as BIO tag.

    The expected format is "[BIO]-TYPE", where TYPE is any non-empty
    nonspace string, and the "-TYPE" part is optional.

    Args:
        tag (string): tag to parse.
    Returns:
        string pair: tag ("B", "I" or "O") and TYPE.
    """

    m = re.match(r'^([BIO])((?:-\S+)?)$', tag)
    assert m, "ERROR: failed to parse tag '%s'" % tag
    ttag, ttype = m.groups()

    # Strip off starting "-" from tagged type, if any.
    if len(ttype) > 0 and ttype[0] == "-":
        ttype = ttype[1:]

    return ttag, ttype

BIO_TAG_RE = re.compile(r'^([BIO])((?:-\S+)?)$')

def is_BIO_tag(s):
    """Return True if given string is a valid BIO tag."""
    return BIO_TAG_RE.match(s)

EMPTY_LINE_RE = re.compile(r'^\s*$')

def _line_is_empty(l):
    return EMPTY_LINE_RE.match(l)

def parse_conll(input, filename=None, separator='\t', is_empty=_line_is_empty):
    """Parse CoNLL-style input.

    Input should consist of blocks of lines separated by empty lines
    (is_empty), each non-empty line consisting of fields separated by
    the given separator.

    Returns:
        list of lists of lists: blocks, lines, fields.
    """

    li, l = 0, None
    try:
        blocks = []
        current_block = []
        for l in input:
            l = l.rstrip()
            li += 1
            if is_empty(l):
                blocks.append(current_block)
                current_block = []
            else:
                current_block.append(l.split(separator))
    except Exception:
        # whatever goes wrong
        raise ParseError(l, li)

    return blocks

def process(fn):
    global token_tag_freq

    with open(fn, 'rU') as f:
        blocks = parse_conll(f)

    for block in blocks:
        for fields in block:
            token, tag = fields[options.tokenindex], fields[options.tagindex]
            
            token_tag_freq[token][tag] += 1

            if token not in first_tag:
                first_tag[token] = tag

def report():
    global options, token_tag_freq

    # simple frequency
#     for token in token_tag_freq:
#         for tag in token_tag_freq[token]:
#             print '%d\t%s\t%s' % (token_tag_freq[token][tag], token, tag)

    # three features for each token: most frequent tag, most frequent
    # non-"O" tag, and first tag
    for token in token_tag_freq:
        # sort by tag frequency
        tag_freq = [(b,a) for a,b in token_tag_freq[token].items()]
        tag_freq.sort(lambda a,b: cmp(b,a))

        mft = tag_freq[0][1]
        mft_nono = "O"
        for freq, tag in tag_freq:
            if tag != "O":
                mft_nono = tag
                break

        if (options.skipout and mft == 'O' and 
            mft_nono == 'O' and first_tag[token] == 'O'):
            continue        

        print '%s\t%s\t%s\t%s' % (token, mft, mft_nono, first_tag[token])
    

def main(argv=None):
    if argv is None:
        argv = sys.argv

    global options

    options = argparser().parse_args(argv[1:])

    try:
        options.tagindex = int(options.tagindex)
    except:
        print >> sys.stderr, 'Failed to parse -i argument "%s" as integer' % options.tagindex
        argparser().print_help()
        return 1
    try:
        options.tokenindex = int(options.tokenindex)
    except:
        print >> sys.stderr, 'Failed to parse -i argument "%s" as integer' % options.tokenindex
        argparser().print_help()
        return 1


    for fn in options.files:
        process(fn)

    report()

if __name__ == "__main__":
    try:
        sys.exit(main(sys.argv))
    except Exception, e:
        print >> sys.stderr, e
        sys.exit(1)
