#!/usr/bin/env python

from subprocess import Popen, PIPE
from os import path

from docproc import Processor
import common

NERSUITE_GTAG_CMD = 'nersuite_gtagger'
NERSUITE_DTAG_CMD = 'nersuite_dic_tagger'
NERSUITE_GTAG_MODEL_DIR = path.join(path.dirname(__file__),
                                    '../models/gtagger/')

class ExternalTagger(Processor):
    def __init__(self):
        Processor.__init__(self)
        self.processes = []

    def _process(self, document):
        """Run external tagger to tag Document.

        Output is added to Token feature vectors.
        """
        
        # To avoid potential buffering issues with large files,
        # process a sentence at a time.
        for s in document.sentences:
            self._tag_sentence(s)

    def _tag_sentence(self, sentence):
        """Run external tagger to tag Sentence.

        Output is added to Token feature vectors.
        """

        input_ = ''.join(sentence.to_nersuite(exclude_tag=True))
        output_lines = self._communicate(input_.encode('utf-8'))
        self._add_tags(sentence, output_lines)

    def _communicate(self, input_):
        assert self.processes, 'cannot tag: no process running'

        # Note: assumes that input and output is one line per token
        # and that sentences are delimited by blank lines. Tagging
        # will fail with assert or deadlock if these assumptions do
        # not hold.

        # exceptional case: empty input -> empty output
        if not input_:
            return []

        input_lines = input_.count('\n')
        assert input_[-2:] in ('\n', '\n\n'), 'missing blank'

        self.processes[0].stdin.write(input_)
        self.processes[0].stdin.flush()

        response = []
        while True:
            l = self.processes[-1].stdout.readline().decode('utf-8')

            response.append(l)
            if not l or l.isspace():
                break
            # fail rather than deadlock on tagger misbehavior
            assert len(response) < input_lines, 'input/output mismatch'

        assert len(response) == input_lines, 'input/output mismatch ' \
            '(%d lines in input, %d in output: %s)' % \
            (input_lines, len(response), ' '.join(response))

        return response

    @classmethod
    def _parse_tagger_line(cls, line):
        fields = line.split('\t')
        start, end, ttext = fields[0], fields[1], fields[2]
        start, end = int(start), int(end)
        features = fields[3:]
        return start, end, ttext, features

    @classmethod
    def _add_tags(cls, sentence, tagger_output):
        # assumes output format of nersuite tagger to match with tokens.
        # TODO: generalize; much of this should go into tagger-specific
        # subclasses instead.

        # exceptional case for empty "sentences"
        if not sentence.tokens:
            assert not tagger_output
            return
        
        # format detail
        assert tagger_output[-1] == '\n'
        assert len(sentence) == len(tagger_output)-1

        for i, token in enumerate(sentence.tokens):
            line = tagger_output[i].rstrip('\n')
            start, end, ttext, features = cls._parse_tagger_line(line)
            
            assert (token.start == start and token.end == end and 
                    token.text == ttext and 
                    token.fvec == features[:len(token.fvec)]), \
                    'input/output mismatch'

            token.fvec.extend(features[len(token.fvec):])

    def stop(self, wait=True):
        """Stop external tagger.
        
        If wait is True, wait for processes to complete; otherwise,
        kill the processes immediately."""

        Processor.stop(self)
        if wait:
            for p in self.processes:
                p.stdin.close()
                p.wait()
        else:
            for p in self.processes:
                p.kill()
        self.processes = []

class GTagger(ExternalTagger):
    def __init__(self, command=NERSUITE_GTAG_CMD):
        ExternalTagger.__init__(self)
        self.command = command

    def start(self):
        ExternalTagger.start(self)
        cmd = [self.command, '-d', NERSUITE_GTAG_MODEL_DIR]
        self.processes.append(Popen(cmd, stdin=PIPE, stdout=PIPE, bufsize=1))

class DictTagger(ExternalTagger):
    def __init__(self, dictionary, parameters=None, command=None):
        ExternalTagger.__init__(self)
        
        if parameters is None:
            parameters = []
        if command is None:
            command = NERSUITE_DTAG_CMD

        self.parameters = parameters
        self.dictionary = dictionary
        self.command = command

    def start(self):
        ExternalTagger.start(self)
        cmd = [self.command] + self.parameters + [self.dictionary]
        self.processes.append(Popen(cmd, stdin=PIPE, stdout=PIPE, bufsize=1))

    @classmethod
    def from_CLI_argument(cls, arg):
        """Initialize and return DictTagger with settings based on
        command-line options."""

        # optional normalization settings separated by colon (':'),
        # optional POS filter settings separated by (additional) colon.

        parameters = []
        if ':' in arg:
            arg, normopt = arg.split(':', 1)
            if ':' in normopt:
                normopt, posopt = normopt.split(':', 1)
                if posopt:
                    parameters.extend(['-p', posopt])
            # separate overlap resolution "tag all" option, if any
            if 'a' in normopt:
                parameters.extend(['-o', 'all'])
                normopt = normopt.replace('a', '')
            if normopt:
                parameters.extend(['-n', normopt])

        return cls(arg, parameters)
