#!/usr/bin/env python

import re
from os import path
from contextlib import contextmanager
from itertools import izip, tee, chain

from sentencesplit import split_sentences

class FormatError(Exception):
    pass

# Context manager that doesn't do anyting on exit (or otherwise; no-op).
@contextmanager 
def noopmanager(thing):
    yield thing

def txt_for_ann(filename):
    """Return the name of the text (.txt) file expected to correspond
    to the given standoff annotation (.ann) file."""

    if not path.isdir(filename):
        return path.splitext(filename)[0]+'.txt'
    else:
        # no change for output to a directory
        return filename

# adapted from http://docs.python.org/library/itertools.html#recipes
def pairwise(iterable, include_last=False):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    if not include_last:
        return izip(a, b)
    else:
        return izip(a, chain(b, (None, )))

# from http://programmaticallyspeaking.com/split-on-separator-but-keep-the-separator-in-python.html
def split_keep_separator(s, sep='\n'):
    return reduce(lambda acc, elem: acc[:-1] + [acc[-1] + elem] if elem == sep 
                  else acc + [elem], re.split("(%s)" % re.escape(sep), s), [])

def text_to_sentences(text, sentence_split=True):
    """Return list of sentences in given text."""

    if not sentence_split:
        sent = split_keep_separator(text)
    else:
        #sent = split_keep_separator(sentencebreaks_to_newlines(text))
        sent = split_sentences(text)

    assert ''.join(sent).replace('\n', ' ') == text.replace('\n', ' '), \
        'internal error'

    return sent

# NERsuite tokenization: alnum sequences preserved as single tokens,
# rest are single-character tokens.
# single-character token. 
# TODO: Unicode support
TOKENIZATION_REGEX = re.compile(r'([0-9a-zA-Z]+|[^0-9a-zA-Z])')

def sentence_to_tokens(text):
    """Return list of tokens in given sentence using NERsuite tokenization."""

    tok = [t for t in TOKENIZATION_REGEX.split(text) if t]
    assert ''.join(tok) == text
    return tok
