#!/usr/bin/env python
# coding=UTF-8

"""Minimalistic truecasing for English."""

from __future__ import with_statement

import sys
import re

import gtbtokenize

TEST_ARG = '--test'

def argparser():
    import argparse

    ap=argparse.ArgumentParser(description="Truecase sentence-split text.")
#     ap.add_argument("-v", "--verbose", default=False, action="store_true", 
#                     help="Verbose output")    
    ap.add_argument("-s", "--suffix", default=None,
                    help="Output file suffix (output to STDOUT if none)")
    ap.add_argument(TEST_ARG, default=False, action="store_true", 
                    help="Perform self-test and exit")
    ap.add_argument("files", metavar="FILE", nargs='+', 
                    help="Text documents (\"-\" for STDIN).")
    return ap


tc_exceptions = None

WORD_TOKEN_REGEX = re.compile('^[a-zA-Z][a-zA-Z0-9-]*$')

INITIAL_CAP_WORD_REGEX = re.compile('^[A-Z][a-zA-Z0-9-]*$')

# Words that commonly appear without change in sentences in Title Case
NON_TITLECASED_WORDS = [
    # from titlecase.py
    "a",
    "after",
    "al", # in "et al"
    "among",
    "an",
    "and",
    "as",
    "at",
    "between",
    "by",
    "due",
    "de",
    "during",
    "et", # in "et al"
    "following",
    "for",
    "from",
    "in",
    "into",
    "of",
    "on",
    "or",
    "so",
    "the",
    "through",
    "to",
    "under",
    "via",
    "vs",
    "with",
    "without",
    # from titles beginning with initial lowercase characters
    "alpha",
    "alpha-Fetoprotein",
    "bcl-2",
    "beta",
    "beta-Adrenergic",
    "beta-Amyloid",
    "beta-Endorphin",
    "beta-Lactamase",
    "cAMP",
    "cAMP-dependent",
    "cDNA",
    "c-erbB-2",
    "c-fos",
    "c-Fos",
    "cGMP",
    "cis-acting",
    "c-Jun",
    "c-myc",
    "c-Myc",
    "delta",
    "eComment",
    "fMRI",
    "gamma-Aminobutyric",
    "mRNA",
    "mtDNA",
    "mTOR",
    "n-3",
    "p16",
    "p38",
    "p53,",
    "p53",
    "p53-dependent",
    "p53-independent",
    "pH",
    "pH-dependent",
    "tert-Butyl",
    "tRNA",
    "von",
]
#NON_TITLECASED_WORD = { w:1 for w in NON_TITLECASED_WORDS }
NON_TITLECASED_WORD = dict([(w,1) for w in NON_TITLECASED_WORDS])

def has_word_char(s):
    return re.search(r'\w', s, re.UNICODE)

def word_tokens(t):
    return [w for w in t if WORD_TOKEN_REGEX.match(w)]

def initial_capital_words(w):
    return [c for c in w if INITIAL_CAP_WORD_REGEX.match(c)]

def sentence_word_tokens(s):
    return word_tokens(gtbtokenize.tokens(s))

TOKEN_NO_UPPER = 'l'
TOKEN_ALL_UPPER = 'u'
TOKEN_FIRST_UPPER = 'f'
TOKEN_MIXED_CASE = 'm'

SENTENCE_STANDARD_CASE = 's'
SENTENCE_ALL_UPPER = 'u'
SENTENCE_ALL_LOWER = 'l'
SENTENCE_TITLE_CASE = 't'
SENTENCE_MIXED_CASE = 'm'

def casedchar(c):
    return c.isupper() or c.islower()

def cased_split(s):
    if len(s) == 0:
        return s

    parts = []
    prev_cased = casedchar(s[0])
    i = 0
    for j in range(1, len(s)):
        cased = casedchar(s[j])
        if cased != prev_cased:
            parts.append(s[i:j])
            i = j
        prev_cased = cased
    parts.append(s[i:])

    assert ''.join(parts) == s
    return parts

def truecase_token(t):
    """Returns the most frequent capitalization of the given token.
    This corresponds to unigram truecasing under the assumption that
    the given form contains no useful case information (i.e. is from
    an all-lowercase or all-uppercase context).

    >>> truecase_token('NORMAL')
    'normal'

    >>> truecase_token('xml')
    'XML'

    >>> truecase_token('mrna')
    'mRNA'
    """

    global tc_exceptions

    l = t.lower()
    if l in tc_exceptions:
        return tc_exceptions[l]
    else:
        return ''.join([tc_exceptions.get(p, p) for p in cased_split(l)])    

def token_category(t):
    """Given a token, return its category as one of the following:
    'l' : no uppercase
    'u' : all characters uppercase
    'f' : first character (only) uppercase
    'm' : mixed uppercase ("other")

    >>> token_category('lower')
    'l'

    >>> token_category('XML')
    'u'

    >>> token_category('Initial')
    'f'

    >>> token_category('MiXeD')
    'm'

    Note also:

    >>> token_category('.')
    'l'

    >>> token_category('X.')
    'u'
    """

    if t.lower() == t:
        return TOKEN_NO_UPPER
    elif t.upper() == t:
        return TOKEN_ALL_UPPER
    elif t[0].lower() != t[0] and t[1:].lower() == t[1:]:
        return TOKEN_FIRST_UPPER
    else:
        return TOKEN_MIXED_CASE

def is_all_upper(s):
    """Return True if the given sentence is all uppercase, False otherwise."""
    return s.upper() == s and s.lower() != s

def is_all_lower(s):
    """Return True if the given sentence is all lowercase, False otherwise."""
    return s.lower() == s and s.upper() != s

INITIAL_MARKER_RE = re.compile(r'^(\s*[({[]?(?:[a-zA-Z][])}:]|(?:[a-zA-Z]?\d+[a-zA-Z]?|[ivxIVX]+)[])}.:])\s+)(\S.*\n*)$')

def split_initial_marker(s):
    """Split given sentence into initial marker and other text.  For
    example, separates sentence-initial "a)", "1.", and "(iv)".

    >>> split_initial_marker('No marker.')
    ('', 'No marker.')

    >>> split_initial_marker('1. Number marker.')
    ('1. ', 'Number marker.')

    >>> split_initial_marker('(IV) Roman numeral marker.')
    ('(IV) ', 'Roman numeral marker.')

    >>> split_initial_marker('S. cerevisiae "S." is not a marker')
    ('', 'S. cerevisiae "S." is not a marker')
    """
    
    m = INITIAL_MARKER_RE.match(s)
    if not m:
        return ('', s)
    else:
        return m.groups()

def remove_initial_marker(s):
    return split_initial_marker(s)[1]
    
def nonstandard_case_words(s, exclude_first_word=True):
    """Return the word tokens in the given sentence that have some
    other than their most frequent casing."""

    if exclude_first_word:
        s = remove_initial_marker(s)

    w = sentence_word_tokens(s)

    if exclude_first_word:
        w = w[1:]

    x = []
    for t in w:
        if not is_all_lower(t) and truecase_token(t) != t:
            x.append(t)

    return x

def is_standard_case(s):
    """Return True if no other than the first word of the given
    sentence has nonstandard capitalization, False otherwise.

    is_standard_case('Normal sentence.')
    True

    is_standard_case('Sentence in Title Case')
    False

    is_standard_case('A) Normal with initial marker.')
    True
    """

    return len(nonstandard_case_words(s)) == 0

def is_title_case(s):
    """
    Return true if the given sentence is in title case, False otherwise.

    Some moderately difficult cases:

    >>> is_title_case('Faust and Moses.')
    False

    >>> is_title_case("The Last Entry in Gauss's Diary.")
    True

    >>> is_title_case('On Interposition as a Cue for Perception of Distance.')
    True

    >>> is_title_case('Geriatrics in Northern Ireland.')
    False

    >>> is_title_case('Effect of Heat Shock on mRNA')
    True

    >>> is_title_case('Portable pH Meter')
    True

    >>> is_title_case('ALL UPPER')
    False

    >>> is_title_case('A) Markered')
    False
    """

    w = sentence_word_tokens(remove_initial_marker(s))

    # algorithm:
    #
    # 1) exclude first word, normally capitalized even if the sentence is
    #    not in title case
    # 2) exclude words that commonly appear without capitalization
    #    even if the sentence is in title case (NON_TITLECASED_WORD)
    # 3) exclude words that are not lowercased but still match their most
    #    frequent capitalization (truecase_token); these are not expected 
    #    to show change in Title Case sentences.
    # 4) exclude all-uppercase words such as "XML"; these cannot inform a
    #    decision in whether a sentence is in Title Case.
    #
    # if there's nothing left (all "word" tokens have expected
    # capitalization both in neutral and Title Case), the sentence
    # should not be judged as Title Case.
    #
    # otherwise assign Title Case if all remaining words have initial caps.

    w = w[1:]

    w = [t for t in w if t not in NON_TITLECASED_WORD]

    w = [t for t in w if not (t.lower() != t and truecase_token(t) == t)]

    w = [t for t in w if not is_all_upper(t)]

    if not w:
        return False

    c = initial_capital_words(w)

    # TODO: consider other alternatives; at the other extreme, having
    # one initial cap word could be taken to indicate Title Case.
    # Debug output for testing alternate heuristics:
#     if len(c) > 0 and len(c) < len(w):
#         print >> sys.stderr, len(w) - len(c), "maybe Title Case:", s, "---", c, "---",  "(", ' '.join([x for x in w if x not in c]), ")"

    return len(c) == len(w)

def sentence_category(s):
    """
    >>> sentence_category('ALL CAPS')
    'u'

    >>> sentence_category('all lower')
    'l'

    >>> sentence_category('First work capitalized')
    's'

    >>> sentence_category('Nouns and Verbs Capitalized and so on.')
    't'
    """

    if is_all_upper(s):
        return SENTENCE_ALL_UPPER
    elif is_all_lower(s):
        return SENTENCE_ALL_LOWER
    elif is_title_case(s):
        return SENTENCE_TITLE_CASE
    elif is_standard_case(s):
        return SENTENCE_STANDARD_CASE
    else:
        return SENTENCE_MIXED_CASE

def truecase_initial(t):
    """Return a version of the given token heuristically restored to
    its "true" case under the assumption that the given form appears
    in sentence-initial position.

    >>> truecase_initial('Normal')
    'normal'

    >>> truecase_initial('XML')
    'XML'

    >>> truecase_initial('January')
    'January'

    >>> truecase_initial('A')
    'a'
    """

    tc = token_category(t) 

    if tc == TOKEN_FIRST_UPPER or (tc == TOKEN_ALL_UPPER and len(t) == 1):
        if truecase_token(t) == t:
            return t
        else:
            return t[0].lower()+t[1:]
    else:
        # no change for all-lower, all-upper, or mixed case.
        return t

def truecase_first_word(s):
    m, r = split_initial_marker(s)
    t = gtbtokenize.tokens(r)
    for i in range(len(t)):
        if has_word_char(t[i]):
            t[i] = truecase_initial(t[i])
            break
    return m + ''.join(t)

def subsentence_split(s):
    """Return a list containing "sub-sentence" components of the given
    sentence with potentially differing casing. 

    Split guarantees that ''.join(subsentence_split(s)) == s

    >>> subsentence_split('RESULTS: The method works.')
    ['RESULTS: ', 'The method works.']

    >>> subsentence_split('NOTE: Space \\t is preserved.\\n')
    ['NOTE: ', 'Space \\t is preserved.\\n']
    """

    # TODO: split out quoted strings from inputs such as
    # 'The book titled "The Health of Nations"'

    # TODO: don't break in parens (e.g. "(Friedman et al. 1993)")

    # *very* simple heuristic, but this might be enough.
    orig_s = s
    components = []
    while True:
        m = re.match(r'^(.*?[.!?:;]\s+)(\S.*\n*)$', s)
        if not m:
            components.append(s)
            break
        else:
            ss, s = m.groups()
            components.append(ss)

    assert ''.join(components) == orig_s, 'subsentence_split error():\n' \
        'orig : "%s"\nsplit: "%s"' % (orig_s, ''.join(components))

    return components

def _truecase_subsentence(s):
    # implementation for truecase()

    sc = sentence_category(s)
    t = gtbtokenize.tokens(s)

    if sc == SENTENCE_STANDARD_CASE:
        return truecase_first_word(s)

    elif sc in (SENTENCE_ALL_UPPER, SENTENCE_ALL_LOWER):
        # no useful case information; use most common form for all tokens.
        t = [truecase_token(w) for w in gtbtokenize.tokens(s)]

    elif sc == SENTENCE_TITLE_CASE:
        # Title Case, truecase initial characters
        t = [truecase_initial(w) for w in gtbtokenize.tokens(s)]
    
    elif sc == SENTENCE_MIXED_CASE:
        # just assume it's standard sentence case (TODO: better processing)
        return truecase_first_word(s)

    else:
        assert False, "INTERNAL ERROR"
        return s

    return ''.join(t)

def truecase(s):
    """Return a version of the given sentence where words are
    heuristically restored to their "true" case, removing
    sentence-initial, Title Case, and ALL CAPS capitalization.

    >>> truecase('Sentence initial capital')
    'sentence initial capital'

    >>> truecase('["Initial cap with preceding nonalpha"]')
    '["initial cap with preceding nonalpha"]'

    >>> truecase(' Space is   preserved,  including    multiple. ')
    ' space is   preserved,  including    multiple. '

    >>> truecase('XML and similar abbreviations remain unchanged.')
    'XML and similar abbreviations remain unchanged.'

    >>> truecase('Sentences in Title Case Processed.')
    'sentences in title case processed.'

    >>> truecase('AS ARE SENTENCES IN ALL CAPS; NOTE: XML')
    'as are sentences in all caps; note: XML'

    >>> truecase('(A) Initial markers are ignored.')
    '(A) initial markers are ignored.'

    >>> truecase('Title-Case Sentences with Hyphens Processed.')
    'title-case sentences with hyphens processed.'

    >>> truecase('IV) Previously buggy case.\\n')
    'IV) previously buggy case.\\n'
    """

    return ''.join([_truecase_subsentence(ss) for ss in subsentence_split(s)])

def read_tc_data():
    global tc_exceptions

    # don't read twice
    if tc_exceptions is not None:
        return

    from os import path
    datafn = path.join(path.abspath(path.dirname(__file__)), 'tcdata.txt')
    with open(datafn, 'rU') as f:
        #tc_exceptions = { t.lower(): t for t in f.read().split() }    
        tc_exceptions = dict([(t.lower(), t) for t in f.read().split()])

def test():
    import doctest
    read_tc_data()
    doctest.testmod()

def process(f, fn, out):
    for l in f:
        l = l.rstrip('\n')
        print >> out, truecase(l)

def main(argv=None):
    global truecase_options

    if argv is None:
        argv = sys.argv

    if TEST_ARG in argv:
        test()
        return 0
    truecase_options = argparser().parse_args(argv[1:])

    read_tc_data()

    for fn in truecase_options.files:        
        if fn == '-':
            process(sys.stdin, '<STDIN>', sys.stdout)
        else:
            with open(fn, 'rU') as f:
                if truecase_options.suffix is None:
                    process(f, fn, sys.stdout)
                else:
                    with open(fn+'.'+truecase_options.suffix, 'wt') as of:
                        process(f, fn, of)

if __name__ == "__main__":
    sys.exit(main(sys.argv))
