#!/usr/bin/env python

import sys
import re

TAG_START_RE = re.compile(r'^\[([^\[\]\s]+)$') # don't match [], [[, etc.
TAG_END_RE = re.compile(r'^(\S+)\]$')

def parse_bracketed(f, fn=None):
    """Parse input in the bracketed format used by the Illinois NER
    tagger from given file-like object. Return (tokens, tags) where
    tokens is a list of token strings and tags a list of tagged spans
    represented as (start, end, type) tuples where start and end are
    indices into tokens.
    """

    tokens = []
    tags = []
    start, type_ = None, None
    for l in f:
        for t in l.split():
            ms = TAG_START_RE.match(t)
            me = TAG_END_RE.match(t)
            if ms:
                assert type_ is None, 'error in %s: %s' % (fn, t)
                type_ = ms.group(1)
                start = len(tokens)
            elif me:
                assert type_ is not None, 'error in %s: %s' % (fn, t)
                tags.append((start, len(tokens), type_))
                token = me.group(1)
                tokens.append(token)
                start, type_ = None, None
            else:
                tokens.append(t)
    return tokens, tags

def parse_bracketed_file(fn):
    with open(fn, 'rU') as f:
        return parse_bracketed(f, fn)

def unescape(t, text, o):
    """Revert some escapes applied by the illinois tagger."""

    origt = t

    if text[o:o+len(t)] == t:
        pass
    elif t == '"' and text[o:o+2] == "''":
        t = "''"
    elif t == '"' and text[o:o+2] == "``":
        t = "``"
    elif text[o:o+len(t)].replace('`', "'") == t:
        t = text[o:o+len(t)]
    else:
        print >> sys.stderr, 'warning: failed to align "%s" with "%s[...]"' %\
            (t, text[o:o+len(t)+10])

    if t != origt:
        print >> sys.stderr, 'note: unescape: %s -> %s' % (origt, t)

    return t

def align_tokens(tokens, text):
    """Align given tokens with given text and return list of start
    offsets for each token."""

    # implementation assumes that the catenation of the tokens differs
    # from the text only in specific escapes and in that the text may
    # have additional space characters.

    # assert ''.join(tokens) == re.sub(r'\s+', '', text), 'error "%s" != "%s"' %\
    #     (''.join(tokens), re.sub(r'\s+', '', text))

    o, offsets = 0, []
    for t in tokens:
        while o < len(text) and text[o].isspace():
            o += 1
        t = unescape(t, text, o)
        assert text[o:o+len(t)] == t, 'error: "%s" != "%s"' % (text[o:o+len(t)], t)
        offsets.append(o)
        o += len(t)

    return offsets

def output_standoff(text, tokens, tags, fn=None, out=sys.stdout):
    try:
        offsets = align_tokens(tokens, text)
    except Exception:
        print >> sys.stderr, 'align_tokens error in %s' % fn
        raise

    i = 1
    for start, end, type_ in tags:
        so, eo = offsets[start], offsets[end]+len(tokens[end])
        ttext = text[so:eo]
        assert ''.join(tokens[start:end+1]) == re.sub(r'\s+', '', ttext)
        print >> out, 'T%d\t%s %s %s\t%s' % (i, type_, so, eo, ttext)
        i += 1

def main(argv=None):
    if argv is None:
        argv = sys.argv

    if len(argv) < 3:
        print >> sys.stderr, 'Usage:', argv[0], 'TAGGED TEXT'
        return 1
    taggedfn, textfn = argv[1:]

    with open(textfn, 'rU') as f:
        text = f.read()
    tokens, tags = parse_bracketed_file(taggedfn)

    output_standoff(text, tokens, tags, taggedfn)

    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv))
