#!/usr/bin/env python

# Script for selecting the best parameters from eval output.

import sys
import re

from os import path
from collections import defaultdict, namedtuple

sys.path.append(path.join(path.dirname(__file__), '../../common'))

# http://stackoverflow.com/a/652226
def makedict():
    return defaultdict(makedict)

RESULT_RE = re.compile(r'^(\S+): p ([0-9.]+)\% \((\d+)/(\d+)\) r ([0-9.]+)\% \((\d+)/(\d+)\) f ([0-9.]+)\%\s*$')

CRITERIA = ['exact', 'left', 'right', 'overlap']

# allowed difference between approximate results in input and exact
# results
EPSILON = 0.1

class ParseError(Exception):
    pass

options = None

def argparser():
    import argparse

    ap=argparse.ArgumentParser()
    ap.add_argument('-c', '--count', default=1, type=int,
                    help='Number of top-ranking parameters to display')
    ap.add_argument('-m', '--macroavg', default=False, action='store_true',
                    help='Use macroaverage (default micro)')
    ap.add_argument('-e', '--exact', default=False, action='store_true',
                    help='Exact matching criterion')
    ap.add_argument('-r', '--right', default=False, action='store_true',
                    help='Right boundary matching criterion')
    ap.add_argument('-l', '--left', default=False, action='store_true',
                    help='Right boundary matching criterion')
    ap.add_argument('-o', '--overlap', default=False, action='store_true', 
                    help='Overlap matching criterion')
    ap.add_argument('-t', '--types', default=None, metavar='TYPE[,...]',
                    help='Types to evaluate for (default TOTAL)')
    ap.add_argument('file', metavar='FILE', nargs=1,
                    help='input data ("-" for STDIN)')
    return ap

def precision(TP, FP):
    if TP + FP == 0:
        return 0.0
    else:
        return 100.0*TP/(TP+FP)    

def recall(TP, FN):
    if TP + FN == 0:
        return 0.0
    else:
        return 100.0*TP/(TP+FN)

def F(p, r):
    if p + r == 0.0:
        return 0.0
    else:
        return 2.0*p*r/(p+r)

Params = namedtuple('Params', ['c2', 'b', 'type_', 'criterion'])

def params_to_str(p):
    return 'c2 %s\tb %s\t%s\t%s:' % (p.c2, p.b, p.type_, p.criterion)

class Result(object):
    def __init__(self, params, TPp, FP, TPg, FN):
        self._params = params
        self.TPp = TPp
        self.TPg = TPg
        self.FP = FP
        self.FN = FN

    @classmethod
    def from_input(cls, params, prec, TPp, predicted, rec, TPg, gold, F_):
        FP = predicted - TPp
        FN = gold - TPg

        # cross-check
        error = precision(TPp, FP) - prec
        if error > EPSILON:
            raise ParseError('Precision mismatch %f: %f vs. %f (%d/%d)' % \
                                 (error, self.precision, self.precision(),
                                  self.TPp, self.TPp+self.FP))
        error = recall(TPg, FN) - rec
        if error > EPSILON:
            raise ParseError('Recall mismatch %f: %f vs. %f (%d/%d)' % \
                                 (error, self.recall, self.recall(),
                                  self.TPg, self.TPg+self.FN))
        error = F(precision(TPp, FP), recall(TPg, FN)) - F_
        if error > EPSILON:
            raise ParseError('F-score mismatch %f: %f vs. %f' % \
                                 (error, self.F, self.F()))

        return cls(params, TPp, FP, TPg, FN)

    def precision(self):
        return precision(self.TPp, self.FP)

    def recall(self):
        return recall(self.TPg, self.FN)

    def F(self):
        return F(self.precision(), self.recall())

    def params(self):
        return self._params

    def __str__(self):
        p = self._params
        return 'c2 %s\tb %s\t%s\t%s: p %.2f (%d/%d) r %.2f (%d/%d) f %.2f' % \
            (p.c2, p.b, p.type_, p.criterion, 
             self.precision(), self.TPp, self.TPp+self.FP,
             self.recall(), self.TPg, self.TPg+self.FN, self.F())

    def __repr__(self):
        return str(self)

    def __cmp__(self, rhs):
        return cmp(self.F(), rhs.F())

def parse_line(l):
    # expected format: lines with TAB-separates fields (C2, B, TYPE,
    # EXACT, LEFT, RIGHT, OVERLAP) where the last four fields that
    # contain results are formatted as in the example "p 35.18%
    # (70/199) r 7.83% (70/894) f 12.81%"

    fields = l.split('\t')
    if len(fields) != 7:
        raise ParseError('Failed to parse line %s' % l)

    # parse results further
    c2, b, type_ = fields[:3]
    c2 = re.sub(r'^c2 ', '', c2)
    b = re.sub(r'^b ', '', b)
    results = fields[3:]
    parsed = []
    for i, r in enumerate(results):
        m = RESULT_RE.match(r)
        if not m:
            raise ParseError('Failed to parse result %s on line %s' % (r, l))
        criterion, precision, TPp, predicted, recall, TPg, gold, F, = m.groups()
        if criterion != CRITERIA[i]:
            raise ParseError('Expected %s, got %s in %s' % (criterion,
                                                            CRITERIA[i], r))
        try:
            precision, recall, F = float(precision), float(recall), float(F)
            TPp, predicted, TPg, gold = int(TPp), int(predicted), int(TPg), \
                int(gold)
        except ValueError:
            raise ParseError('Failed to value in result %s on line %s' % (r, l))

        result = Result.from_input(Params(c2, b, type_, criterion),
                                   precision, TPp, predicted, 
                                   recall, TPg, gold, F)
        parsed.append(result)

    return parsed

def parse_lines(f):
    all_results = []
    for l in f:
        l = l.rstrip()
        if not l or l.isspace():
            continue
        all_results.extend(parse_line(l))        
    return all_results

def parse(fn):
    if fn == '-':
        results = parse_lines(sys.stdin)        
    else:
        try:
            with open(fn, 'rU') as f:
                results = parse_lines(f)
        except IOError, e:
            print e
            return None
    return results

def include_result(result, criterion):
    params = result.params()
    if params.type_ not in options.types:
        return False
    if not params.criterion == criterion:
        return False
    return True

def filter_results(results, criterion):
    filtered = []
    for r in results:
        if include_result(r, criterion):
            filtered.append(r)
    return filtered

def average_result(results, macroavg):
    # should only ever average over results with identical parameters
    params = results[0].params()
    assert all(r.params() == params for r in results), 'Mismatched params'
    if macroavg:
        return (sum([r.F() for r in results])/len(results), params)
    else:
        TPp = sum([r.TPp for r in results])
        FP = sum([r.FP for r in results])
        TPg = sum([r.TPg for r in results])
        FN = sum([r.FN for r in results])
        return (Result(params, TPp, FP, TPg, FN), params)

def performance_rank(results, macroavg):
    results_by_params = defaultdict(list)
    performance_by_params = {}
    for r in results:
        results_by_params[r.params()].append(r)
    average = []
    for p in results_by_params:
        average.append(average_result(results_by_params[p], macroavg))
    average.sort(lambda a,b: cmp(b,a))
    return average

def process(fn):
    global options

    results = parse(fn)
    if results is None:
        return 1
    for i in range(options.count):
        if options.count != 1:
            print 'rank %d' % (i+1)
        for c in [c for c in CRITERIA if vars(options)[c]]:
            filtered = filter_results(results, c)
            ranked = performance_rank(filtered, options.macroavg)
            for r in ranked[i:i+1]:
                if isinstance(r[0], Result):
                    print r[0]
                else:
                    print '%s f %.2f' % (params_to_str(r[1]), r[0])

def main(argv):
    global options
    options = argparser().parse_args(argv[1:])

    if options.count < 1:
        print >> sys.stderr, 'error: --count must be >= 1'
        return 1

    # use all matching criteria by default
    if (not options.exact and not options.overlap and
        not options.left and not options.right):
        options.exact = options.left = options.right = options.overlap = True

    # look at TOTAL lines by default
    if options.types is None:
        options.types = ['TOTAL']
    else:
        options.types = options.types.split(',')

    for fn in options.file:
        process(fn)
        
    return 0

if __name__ == "__main__":
    sys.exit(main(sys.argv))
