#!/usr/bin/env python

# Filter string-type pairs by occurrence statistics from TAB-separated
# representation of MetaMap output.

# This script expects input in the format output by the script
# mmtsvextract.py

import sys
import re

options = None

def argparser():
    import argparse

    ap=argparse.ArgumentParser(description='Filter string-type pairs by ' +
                               'occurrence statistics from MetaMap output.')
    ap.add_argument('-t', '--total', type=int, default=None, 
                    help='remove strings with total occurrence count < COUNT')
    ap.add_argument('-r', '--ratio', type=float, default=None, 
                    help='remove types with occurrence ratio < RATIO')
    ap.add_argument('-n', '--no-count', default=False, action='store_true', 
                    help='output types only, without counts or ratios.')
    ap.add_argument('-v', '--verbose', default=False, action='store_true', 
                    help='verbose output.')
    ap.add_argument('file', metavar='FILE', nargs=1, 
                    help='Input file.')
    return ap

TOTAL_KEY = '<TOTAL>'

def get_total(fields):
    total = None
    other_fields = []
    for f in fields:
        n, l = f.split(':', 1)
        if l == TOTAL_KEY:
            assert total is None, "Multiple totals"
            total = int(n)
        else:
            other_fields.append(f)
    return total, other_fields

def filter_fields(fields, total):
    global options

    filtered = []
    for f in fields:
        n, l = f.split(':', 1)
        ratio = None

        if n.isdigit():
            # int; need total for ratio
            if total is not None:
                ratio = 1.0*int(n)/total
        else:
            # assume float giving ratio
            ratio = float(n)

        if options.ratio:
            assert ratio is not None, "input is missing ratio"
            if ratio >= options.ratio:
                filtered.append((n,l))
    
    return filtered

def output(out, string, fields, total):
    global options

    out.write('%s' % string)
    if total is not None and not options.no_count:
        out.write('\t%d:%s' % (total, TOTAL_KEY))
    for n, l in fields:
        if options.no_count:
            out.write('\t%s' % l)
        else:
            out.write('\t%s:%s' % (n, l))
    out.write('\n')
    
def process(fn):
    global options

    total_filtered, ratio_filtered, output_count = 0, 0, 0

    with open(fn, 'rU') as f:
        for l in f:
            l = l.rstrip('\n')

            fields = l.split('\t')
            string, fields = fields[0], fields[1:]

            total, fields = get_total(fields)

            if options.total is not None:
                assert total is not None, "input is missing total"
                if total < options.total:
                    total_filtered += 1
                    continue
            
            filtered_fields = filter_fields(fields, total)
            ratio_filtered += len(fields) - len(filtered_fields)
            fields = filtered_fields

            if not fields:
                continue

            output(sys.stdout, string, fields, total)
            output_count += 1

    total = total_filtered + ratio_filtered + output_count

    if options.verbose:
        for s, c in (('Output', output_count),
                     ('filtered by total', total_filtered),
                     ('filtered by ratio', ratio_filtered)):
            print >> sys.stderr, '%s %d (%.1f%%),' % (s, c, 100.0*c/total),
        print >> sys.stderr
            
def main(argv=None):
    global options

    if argv is None:
        argv = sys.argv

    options = argparser().parse_args(argv[1:])

    if options.ratio is None and options.total is None:
        print >> sys.stderr, 'error: please specify filtering criterion'
        return 1
    if options.total is not None and not options.total > 0:
        print >> sys.stderr, 'error: need total > 0'
        return 1
    if options.ratio is not None and not 0.0 < options.ratio <= 1.0:
        print >> sys.stderr, 'error: need 0.0 < ratio <= 1.0'
        return 1

    process(options.file[0])

if __name__ == '__main__':
    sys.exit(main(sys.argv))
