#!/usr/bin/env python

# Extract string-type statistics from TAB-separated representation of
# MetaMap output.

# This script expects input in the format output by the script
# mmxml2tsv.py.

import sys
import os
import re
import gzip
import decimal

from collections import defaultdict
from itertools import groupby

options = None

def argparser():
    import argparse

    ap=argparse.ArgumentParser(description='Extract string-type statistics '+
                               'from MetaMap output in TSV format.')
    ap.add_argument('-a', '--all', default=False, action='store_true',
                    help='print all strings, including unmapped')
    ap.add_argument('-n', '--nototals', default=False, action='store_true',
                    help='exclude total occurrence count from output')
    ap.add_argument('-r', '--ratios', default=False, action='store_true',
                    help='output ratios instead of counts')
    ap.add_argument('-s', '--sigfig', type=int, default=2,
                    metavar='NUM',
                    help='number of significant digits (with -r)')
    ap.add_argument('-m', '--mapping-threshold', type=int, default=0, 
                    metavar='SCORE',
                    help='ignore mappings w/higher score (range [-1000,0])')
    ap.add_argument('-c', '--candidate-threshold', type=int, default=0, 
                    metavar='SCORE',
                    help='ignore candidates w/higher score (range [-1000,0])')
    ap.add_argument('-v', '--verbose', default=False, action='store_true', 
                    help='verbose output.')
    ap.add_argument('files', metavar='FILE', nargs='+', 
                    help='Input file(s).')
    return ap

PMID_RE = re.compile(r'^\tPMID: (\d+)\s*$')
UTTERANCE_RE = re.compile(r'^\tUtterance ?(\((?:ti|ab)\))?: (.*)$')
PHRASE_RE = re.compile(r'^\tPhrase: (.*)$')

def float_to_decimal(f):
    # from http://docs.python.org/library/decimal.html#decimal-faq
    'Convert a floating point number to a Decimal with no loss of information'
    n, d = f.as_integer_ratio()
    numerator, denominator = decimal.Decimal(n), decimal.Decimal(d)
    ctx = decimal.Context(prec=60)
    result = ctx.divide(numerator, denominator)
    while ctx.flags[decimal.Inexact]:
        ctx.flags[decimal.Inexact] = False
        ctx.prec *= 2
        result = ctx.divide(numerator, denominator)
    return result 

def fformat(number, sigfig):
    # from http://stackoverflow.com/questions/2663612/nicely-representing-a-floating-point-number-in-python/2663623#2663623
    assert sigfig > 0
    try:
        d = decimal.Decimal(number)
    except TypeError:
        d = float_to_decimal(float(number))
    sign, digits, exponent = d.as_tuple()
    if len(digits) < sigfig:
        digits = list(digits)
        digits.extend([0] * (sigfig - len(digits)))    
    shift = d.adjusted()
    result = int(''.join(map(str,digits[:sigfig])))
    # Round the result
    if len(digits) > sigfig and digits[sigfig] >= 5:
        result+=1
    result=list(str(result))
    # Rounding can change the length of result
    # If so, adjust shift
    shift += len(result) - sigfig
    # reset len of result to sigfig
    result = result[:sigfig]
    if shift >= sigfig-1:
        # Tack more zeros on the end
        result+=['0']*(shift-sigfig+1)
    elif 0 <= shift:
        # Place the decimal point in between digits
        result.insert(shift+1,'.')
    else:
        # Tack zeros on the front
        assert shift < 0
        result=['0.']+['0']*(-shift-1)+result
    if sign:
        result.insert(0,'-')
    return ''.join(result)

def read_PMID(l):
    m = PMID_RE.match(l)
    if not m:
        return None
    else:
        return m.group(1)

def read_utterance(l):
    m = UTTERANCE_RE.match(l)
    if not m:
        return None
    else:
        return m.groups()

def read_phrase(l):
    m = PHRASE_RE.match(l)
    if not m:
        return None
    else:
        return m.group(1)

class Document(object):
    def __init__(self, PMID):
        self.PMID = PMID
        self.utterances = []

class Utterance(object):
    def __init__(self, context, text):
        self.context = context
        self.text = text
        self.phrases = []

class Phrase(object):
    def __init__(self, text):
        self.text = text
        self.candidates = []

    def filter_candidates(self, mapping_threshold=0, candidate_threshold=0,
                          best_mapping_only=True):
        # Note: MetaMap output scores are negated, with -1000 being
        # the best possible score and 0 the worst.
        self.candidates = [c for c in self.candidates if
                           c.mapping_score <= mapping_threshold and
                           c.candidate_score <= candidate_threshold]
        if best_mapping_only and self.candidates:
            best_mapping_score = max([c.mapping_score for c in self.candidates])
            self.candidates = [c for c in self.candidates if 
                               c.mapping_score == best_mapping_score]

    def string_type_pairs(self):
        string_type = set()
        for c in self.candidates:
            for t in c.types:
                string_type.add((c.source_text, t))
        return string_type

    def unmapped_strings(self):
        mapped_offset = {}
        for c in self.candidates:
            for start, end in c.offsets:
                for i in range(start, end):
                    mapped_offset[i] = True

        # separate maximal sequences of unmapped characters
        unmapped_chars = [self.text[i] if i not in mapped_offset else None
                          for i in range(len(self.text))]
        unmapped_strings = []
        for grouper in groupby(unmapped_chars, lambda c: c is not None):
            if grouper[0]:
                unmapped_strings.append(''.join(grouper[1]))

        # remove empties and normalize
        unmapped_strings = [normspace(s) for s in unmapped_strings if
                            not s.isspace()]

        return unmapped_strings

class Candidate(object):
    def __init__(self, tsv):
        self.tsv = tsv
        fields = self.tsv.split('\t')
        mapping_score, candidate_score, source_text, types = fields[0:4]

        self.mapping_score = int(mapping_score)
        self.candidate_score = int(candidate_score)
        self.source_text = source_text
        self.types = types.split(',')

        # character offset format is START-END[;START-END[...]]
        self.offsets = []
        for range in fields[8].split(';'):
            start, end = range.split('-')
            self.offsets.append((int(start), int(end)))

def read_document(input):
    current_document = None

    for l in input:
        l = l.rstrip('\n')

        if l[0] == '\t':
            # context line: either PMID, Utterance, or Phrase
            if read_PMID(l) is not None:
                PMID = read_PMID(l)
                if (current_document is not None and 
                    current_document.PMID != PMID):
                    # doc change, have previous
                    yield current_document
                if (current_document is None or 
                    current_document.PMID != PMID):
                    # new doc
                    current_document = Document(PMID)
            elif read_utterance(l) is not None:
                context, text = read_utterance(l)
                utterance = Utterance(context, text)
                current_document.utterances.append(utterance)
            elif read_phrase(l) is not None:
                text = read_phrase(l)
                phrase = Phrase(text)
                current_document.utterances[-1].phrases.append(phrase)
            else:
                assert False, 'Failed to parse line %s' % l
        else:
            # candidate line
            c = Candidate(l)
            current_document.utterances[-1].phrases[-1].candidates.append(c)

    # end of input, yield final doc (if any)
    if current_document is not None:
        yield current_document

def normspace(s):
    return re.sub('\s+', ' ', s.strip())

TOTAL_KEY = '<TOTAL>'

def process(fn, count=None):
    global options

    if options.verbose:
        print >> sys.stderr, 'Processing %s ...' % fn

    if count is None:
        count = defaultdict(lambda: defaultdict(int))

    # wrap zip files
    input = fn
    if os.path.splitext(fn)[1] == '.gz':
        input = gzip.GzipFile(fn)
    elif fn == '-':
        input = sys.stdin
    else:
        input = open(fn, 'rU')

    orig_candidate_count, filtered_candidate_count = 0, 0

    for dn, d in enumerate(read_document(input)):
        for u in d.utterances:
            for p in u.phrases:
                orig_candidate_count += len(p.candidates)
                p.filter_candidates(options.mapping_threshold,
                                    options.candidate_threshold)
                filtered_candidate_count += len(p.candidates)

        # Take counts of both string-type pairs and unmapped strings.
        # TODO: consider two extensions of unmapped string processing:
        # tokenization and substrings. The current implementation
        # simply counts each string as it appears, so that
        # e.g. "thus," does not count as "thus" (w/o comma) and "of
        # the" does not count as either "of" or "the".

        for u in d.utterances:
            for p in u.phrases:
                strings = set()
                for s, t in p.string_type_pairs():
                    count[s][t] += 1
                    strings.add(s)
                for s in strings:
                    count[s][TOTAL_KEY] += 1
                for s in p.unmapped_strings():
                    count[s][None] += 1
                    count[s][TOTAL_KEY] += 1

    if options.verbose:
        print >> sys.stderr, 'Filtered %d candidates to %d (%.1f%%)' % \
            (orig_candidate_count, filtered_candidate_count,
             100.0 * filtered_candidate_count / orig_candidate_count)

    if input != sys.stdin:
        input.close()

    return count

def output(count, out):
    global options

    # output most frequent first
    strings = count.keys()
    strings.sort(lambda a,b: cmp(count[b][TOTAL_KEY], count[a][TOTAL_KEY]))

    for s in strings:
        types = count[s].keys()
        types = [t for t in types if t is not None and t != TOTAL_KEY]

        if not options.all and len(types) == 0:
            continue

        types.sort(lambda a,b: cmp(count[s][b], count[s][a]))

        out.write('%s\t' % s)

        if not options.nototals:
            out.write('%d:%s\t' % (count[s][TOTAL_KEY], TOTAL_KEY))

        c = count[s]
        f = options.sigfig
        if not options.ratios:
            print >> out, '\t'.join(['%d:%s' % (c[t], t) for t in types])
        else:
            total = count[s][TOTAL_KEY]
            print >> out, '\t'.join(['%s:%s' % (fformat(1.0*c[t]/total, f), t)
                                     for t in types])

def main(argv=None):
    global options

    if argv is None:
        argv = sys.argv

    options = argparser().parse_args(argv[1:])

    for t, n in ((options.candidate_threshold, 'Candidate'),
                 (options.mapping_threshold, 'Mapping')):
        if t < -1000 or t > 0:
            print >> sys.stderr, '%s threshold must be between -1000 and 0' % n
            argparser().print_help()
            return 1

    count = None
    for fn in options.files:
        count = process(fn, count)

    output(count, sys.stdout)

    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv))
