#!/usr/bin/env python

# Convert MetaMap XML output into simple TSV representation. Note that
# the conversion is lossy and ignores information not related to the
# primary phrase tagging output.

# The output of this script is line-oriented, with two line types:
#
# 1) context lines begin with a TAB character, followed by a string
#    identifying the type of context information, a ": " separator,
#    and the corresponding information:
#
#  - PMID: PubMed ID of source document
#  - Utterance: string from source document per MetaMap segmentation
#  - Phrase: substring of utterance per MetaMap segmentation
#
# 2) candidate lines do not begin with a TAB, and contain the following
#    TAB-separated fields, representing one mapping candidate concept
#    per line:
#
#  - mapping score (int): score of mapping containing the candidate
#  - candidate score (int): score of the candidate itself
#  - source text: matched text as it appears in context phrase 
#  - semantic types (comma-separated list): UMLS semantic types
#  - concept ID: UMLS concept ID of candidate
#  - matched words: matched text as processed by MetaMap
#  - preferred form: the preferred form of the candidate concept
#  - matched candidate: the matched form of the candidate concept

import sys
import re
import os
import gzip

try:
    import xml.etree.ElementTree as ET
except ImportError: 
    import cElementTree as ET   

options = None

def argparser():
    import argparse

    ap=argparse.ArgumentParser(description="Convert MetaMap output XML into " +
                               "TSV representation.")
    ap.add_argument("-v", "--verbose", default=False, action="store_true", 
                    help="Verbose output.")
    ap.add_argument("files", metavar="FILE", nargs="+", 
                    help="Input file(s).")
    return ap

def normspace(s):
    return re.sub('\s', ' ', s)

class StructureError(Exception):
    pass

def onlychild(element, childtag):
    children = element.findall(childtag)
    
    if len(children) != 1:
        raise StructureError("expected single %s child in <%s> found %d" % 
                             (childtag, element.tag, len(children)))
    return children[0]

class Candidate(object):
    def __init__(self, e):
        """Construct Candidate from MetaMap XML <Candidate> element."""

        # Note: ignores the following <Candidate> subelements:
        # <Sources>, <MatchMaps>, <IsHead>, <IsOverMatch>

        self.score = int(onlychild(e, "CandidateScore").text)
        self.cui = onlychild(e, "CandidateCUI").text
        self.matched = onlychild(e, "CandidateMatched").text
        self.preferred = onlychild(e, "CandidatePreferred").text

        self.matchedwords = []
        for matchedword in onlychild(e, "MatchedWords"):
            assert matchedword.tag == "MatchedWord"
            self.matchedwords.append(matchedword.text)

        self.semtypes = []
        for semtype in onlychild(e, "SemTypes"):
            assert semtype.tag == "SemType"
            self.semtypes.append(semtype.text)

        self.offsets = []
        for conceptpi in onlychild(e, "ConceptPIs"):
            assert conceptpi.tag == "ConceptPI"
            start = int(onlychild(conceptpi, "StartPos").text)
            length = int(onlychild(conceptpi, "Length").text)
            self.offsets.append((start, start+length))
        self.offsets.sort()

        # need information from Mapping to determine source texts and
        # offsets relative to these texts; use settext()
        self.texts = []
        self.reloffsets = []

    def settext(self, text, baseoffset):
        for start, end in self.offsets:
            start, end = start-baseoffset, end-baseoffset
            self.texts.append(text[start:end])
            self.reloffsets.append((start, end))

    def __str__(self):
        return "%d\t%s\t%s\t%s\t%s\t%s\t%s\t%s" % \
            (self.score, normspace(' '.join(self.texts)),
             ','.join(self.semtypes), self.cui, 
             normspace(' '.join(self.matchedwords)),
             normspace(self.preferred), normspace(self.matched),
             ';'.join(['%d-%d' % (s, e) for s, e in self.reloffsets]))
             
class Mapping(object):
    def __init__(self, e):
        """Construct Mapping from MetaMap XML <Mapping> element."""

        self.score = int(onlychild(e, "MappingScore").text)
        self.candidates = []

        candidates = onlychild(e, "MappingCandidates")
        for candidate in candidates:
            if candidate.tag != "Candidate":
                raise StructureError("<MappingCandidates> has non-<Candidate>" +
                                     "child <%s>" % candidate.tag)
            self.candidates.append(Candidate(candidate))

    def settext(self, text, baseoffset):
        for c in self.candidates:
            c.settext(text, baseoffset)
        
    def __str__(self):
        return '\n'.join(['%d\t%s' % (self.score, c) for c in self.candidates])

class Phrase(object):
    def __init__(self, e):
        """Construct Phrase from MetaMap XML <Phrase> element."""
        
        self.text = onlychild(e, "PhraseText").text
        self.start = int(onlychild(e, "PhraseStartPos").text)
        self.length = int(onlychild(e, "PhraseLength").text)
        self.mappings = []

        mappings = onlychild(e, "Mappings")
        for mapping in mappings:
            if mapping.tag != "Mapping":
                raise StructureError("<Mappings> has non-<Mapping> child " +
                                     "<%s>" % mapping.tag)
            self.mappings.append(Mapping(mapping))

        # propagate text and start offset to allow candidates to
        # determine their source texts
        for m in self.mappings:
            m.settext(self.text, self.start)

    def __str__(self):
        return '\n'.join(["\tPhrase: %s" % normspace(self.text)]+
                         [str(m) for m in self.mappings])
        
def process(fn):
    global options

    # wrap zip files
    input = fn
    zipped = False
    if os.path.splitext(fn)[1] == '.gz':
        input = gzip.GzipFile(fn)
        zipped = True
    if fn == '-':
        input = sys.stdin

    for event, element in ET.iterparse(input):
        # process complete utterance elements ("end" event)
        if event != "end" or element.tag != "Utterance":
            continue
        utterance = element

        try:
            print "\tPMID: %s" % onlychild(utterance, "PMID").text
            print "\tUtterance (%s): %s" % \
                (normspace(onlychild(utterance, "UttSection").text),
                 normspace(onlychild(utterance, "UttText").text))

            phrases = onlychild(utterance, "Phrases")
            for phrase in phrases:
                if phrase.tag != 'Phrase':
                    raise StructureError("<Phrases> has non-<Phrase> child " +
                                         "<%s>" % phrase.tag)
                print Phrase(phrase)

            utterance.clear()

        except StructureError, e:
            print >> sys.stderr, "Error:", e

    if zipped:
        input.close()

def main(argv=None):
    global options

    if argv is None:
        argv = sys.argv

    try:
        options = argparser().parse_args(argv[1:])
    except ImportError:
        print >> sys.stderr, "No argparse, processing <STDIN>"
        from collections import namedtuple
        options = namedtuple('Options', 'files')(['-'])

    for fn in options.files:
        process(fn)

    return 0

if __name__ == "__main__":
    sys.exit(main(sys.argv))
